#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class CameraMatrix, DistortionCoefficients, CameraIntrinsics.
"""

import math
from typing import Dict, Generator, Iterator, Mapping, Optional, Sequence, Tuple

import numpy as np

from ..geometry import Vector2D


class CameraMatrix:
    """A class representing camera matrix.

    :param matrix: A 3x3 Sequence of camera matrix
    :param loads: A dict containing parameters of camera matrix
    :param kwargs: Float values which keys must be "fx", "fy", "cx", "cy" and "skew"(optional)
    :raises KeyError: When essential keys are missing in param loads
    :raises TypeError: When kwargs do not have correct keys and only param kwargs is provided
    """

    def __init__(
        self,
        matrix: Optional[Sequence[Sequence[float]]] = None,
        *,
        loads: Optional[Dict[str, float]] = None,
        **kwargs: float,
    ) -> None:
        if loads:
            kwargs = loads

        if kwargs:
            try:
                self._fx = kwargs["fx"]
                self._fy = kwargs["fy"]
                self._cx = kwargs["cx"]
                self._cy = kwargs["cy"]
                self._skew = kwargs.get("skew", 0)
                return
            except KeyError as error:
                if loads:
                    raise
                if matrix is None:
                    raise TypeError(
                        f"Missing key {error} in kwargs to initialize {self.__class__.__name__}"
                    )

        if matrix is not None:
            self._fx = matrix[0][0]
            self._fy = matrix[1][1]
            self._cx = matrix[0][2]
            self._cy = matrix[1][2]
            self._skew = matrix[0][1]
            return

        raise TypeError(f"Require 'fx', 'fy', 'cx', 'cy' to initialize {self.__class__.__name__}")

    def dumps(self) -> Dict[str, float]:
        """Dumps the camera matrix into a dict."""
        return {
            "fx": self._fx,
            "fy": self._fy,
            "cx": self._cx,
            "cy": self._cy,
            "skew": self._skew,
        }

    def __repr__(self) -> str:
        str_list = [
            f"{self.__class__.__name__}(",
            f"  fx={self._fx},",
            f"  fy={self._fy},",
            f"  cx={self._cx},",
            f"  cy={self._cy},",
        ]
        if self._skew:
            str_list.append(f"  skew={self._skew},")
        str_list.append(")")
        return "\n".join(str_list)

    @property
    def fx(self) -> float:  # pylint: disable=invalid-name
        """Get the fx of the camera matrix."""
        return self._fx

    @property
    def fy(self) -> float:  # pylint: disable=invalid-name
        """Get the fy of the camera matrix."""
        return self._fy

    @property
    def cx(self) -> float:  # pylint: disable=invalid-name
        """Get the cx of the camera matrix."""
        return self._cx

    @property
    def cy(self) -> float:  # pylint: disable=invalid-name
        """Get the cy of the camera matrix."""
        return self._cy

    @property
    def skew(self) -> float:
        """Get the skew of the camera matrix."""
        return self._skew

    def as_matrix(self) -> np.ndarray:
        """Returns camera matrix as a 3x3 matrix.

        :return: A 3x3 numpy array representing the camera matrix
        """
        return np.array(
            [
                [self._fx, self._skew, self._cx],
                [0.0, self._cy, self._cy],
                [0.0, 0.0, 1.0],
            ]
        )

    def project(self, point: Sequence[float]) -> Vector2D:
        """Project a point to the pixel coordinate.

        :param point: A Sequence containing coordinates of the point to be projected
        :raises TypeError: When the dimension of input point is not two or three
        """
        # pylint: disable=invalid-name
        if len(point) == 3:
            x = point[0] / point[2]
            y = point[1] / point[2]
        elif len(point) == 2:
            x = point[0]
            y = point[1]
        else:
            raise TypeError("The point to be projected must have 2 or 3 dimension")

        x = self._fx * x + self._skew * y + self._cx
        y = self._fy * y + self._cy
        return Vector2D(x, y)


class DistortionCoefficients(Mapping[str, float]):
    """A class representing camera distortion coefficients.

    :param loads: A dict containig distortion coefficients of a camera
    :param kwargs: Float values which keys must be k1, k2, ... and p1, p2, ...
    """

    _FISHEYE_MINIMUM_R = 1e-8

    def __init__(self, *, loads: Optional[Dict[str, float]] = None, **kwargs: float) -> None:
        self._data: Dict[str, float] = {}
        if loads:
            kwargs = loads

        self._data.update(dict(self._distortion_generator("p", kwargs)))
        self._data.update(dict(self._distortion_generator("k", kwargs)))

        if not self._data:
            raise TypeError(
                f"Require tangential or radial distortion to initialize {self.__class__.__name__}"
            )

    def __repr__(self) -> str:
        str_list = [f"{self.__class__.__name__}{{"]
        for key, value in self._data.items():
            str_list.append(f'  "{key}": {value},')
        str_list.append("}")
        return "\n".join(str_list)

    def __getitem__(self, key: str) -> float:
        return self._data.__getitem__(key)

    def __len__(self) -> int:
        return self._data.__len__()

    def __iter__(self) -> Iterator[str]:
        return self._data.__iter__()

    def dumps(self) -> Dict[str, float]:
        """Dumps the distortion coefficients into a dict.

        :return: A dict containing all the camera distortion coefficients
        """
        return self._data.copy()

    def distort(self, point: Sequence[float], is_fisheye: bool = False) -> Vector2D:
        """Add distortion a point.

        :param point: A Sequence containing coordinates of the point to be distorted
        :raises TypeError: When the dimension of input point is not two or three
        :return: Distorted 2D point
        """
        # pylint: disable=invalid-name
        if len(point) == 3:
            x = point[0] / point[2]
            y = point[1] / point[2]
        elif len(point) == 2:
            x = point[0]
            y = point[1]
        else:
            raise TypeError("The point to be projected must have 2 or 3 dimension")

        x2 = x ** 2
        y2 = y ** 2
        xy2 = 2 * x * y
        r2 = x2 + y2

        radial_distortion = self._calculate_radial_distortion(r2, is_fisheye)
        tangential_distortion = self._calculate_tangential_distortion(r2, x2, y2, xy2, is_fisheye)
        x = x * radial_distortion + tangential_distortion[0]
        y = y * radial_distortion + tangential_distortion[1]
        return Vector2D(x, y)

    def _calculate_radial_distortion(self, r2: float, is_fisheye: bool = False) -> float:
        # pylint: disable=invalid-name
        if is_fisheye:
            r = math.sqrt(r2)
            factor = math.atan(r)
            factor2 = factor ** 2
        else:
            factor2 = r2

        radial_distortion = 1.0
        for i, (_, value) in enumerate(self._distortion_generator("k", self._data), 1):
            radial_distortion += value * factor2 ** i

        if is_fisheye:
            radial_distortion = radial_distortion * factor / r if r > self._FISHEYE_MINIMUM_R else 1
        return radial_distortion

    def _calculate_tangential_distortion(  # pylint: disable=too-many-arguments
        self, r2: float, x2: float, y2: float, xy2: float, is_fisheye: bool
    ) -> Tuple[float, float]:
        # pylint: disable=invalid-name
        if is_fisheye:
            return (0, 0)

        p1 = self._data["p1"]
        p2 = self._data["p2"]
        return (p1 * xy2 + p2 * (r2 + 2 * x2), p1 * (r2 + 2 * y2) + p2 * xy2)

    @staticmethod
    def _distortion_generator(
        distortion_keyword: str, data: Dict[str, float]
    ) -> Generator[Tuple[str, float], None, None]:
        for index in range(1, len(data) + 1):
            key = f"{distortion_keyword}{index}"
            if key not in data:
                break
            yield (key, data[key])


class CameraIntrinsics:
    """A class representing camera intrinsics including camera matrix and distortion coeffecients.

    :param camera_matrix: A 3x3 Sequence of camera matrix
    :param loads: A dictionary containig 'cameraMatrix' and 'distortionCoefficients'
    :param kwargs: Float values to initialize CameraMatrix and DistortionCoefficients
    """

    def __init__(
        self,
        camera_matrix: Optional[Sequence[Sequence[float]]] = None,
        *,
        loads: Optional[Dict[str, Dict[str, float]]] = None,
        _init_distortion: bool = True,
        **kwargs: float,
    ) -> None:
        if loads:
            self._camera_matrix = CameraMatrix(loads=loads["cameraMatrix"])
            if "distortionCoefficients" not in loads:
                self._distortion_coefficients = None
            else:
                self._distortion_coefficients = DistortionCoefficients(
                    loads=loads["distortionCoefficients"]
                )
            return

        self._camera_matrix = CameraMatrix(camera_matrix, loads=None, **kwargs)
        if kwargs and _init_distortion:
            try:
                self._distortion_coefficients = DistortionCoefficients(loads=None, **kwargs)
                return
            except TypeError:
                pass
        self._distortion_coefficients = None

    def dumps(self) -> Dict[str, Dict[str, float]]:
        """Dumps the camera intrinsics into a dictionary."""
        data = {"cameraMatrix": self._camera_matrix.dumps()}
        if self._distortion_coefficients:
            data["distortionCoefficients"] = self._distortion_coefficients.dumps()

        return data

    def __repr__(self) -> str:
        str_list = [
            f"{self.__class__.__name__}(",
            "  CameraMatrix(",
            f"    fx={self._camera_matrix.fx},",
            f"    fy={self._camera_matrix.fy},",
            f"    cx={self._camera_matrix.cx},",
            f"    cy={self._camera_matrix.cy},",
        ]
        if self._camera_matrix.skew:
            str_list.append(f"    skew={self._camera_matrix.skew},")
        str_list.append("  )")

        if self._distortion_coefficients:
            str_list.append("  DistortionCoefficients{")
            for key, value in self._distortion_coefficients.items():  # pylint: disable=no-member
                str_list.append(f'    "{key}": {value},')
            str_list.append("  }")

        str_list.append(")")
        return "\n".join(str_list)

    @property
    def camera_matrix(self) -> CameraMatrix:
        """Get the camera matrix of camera intrinsics.

        :return: CameraMatrix class object containing fx, fy, cx, cy, skew(optional)
        """
        return self._camera_matrix

    @property
    def distortion_coefficients(self) -> Optional[DistortionCoefficients]:
        """Get the distortion coefficients of camera intrinsics, could be None.

        :return: DistortionCoefficients class object
                containing tangential and radial distortion coefficients
        """
        return self._distortion_coefficients

    def set_camera_matrix(
        self,
        matrix: Optional[Sequence[Sequence[float]]] = None,
        **kwargs: float,
    ) -> None:
        """Set camera matrix of camera intrinsics.

        :param matrix: Camera matrix in 3x3 sequence
        :param kwargs: Keyword arguments containing fx, fy, cx, cy, skew(optional)
        """
        self._camera_matrix = CameraMatrix(matrix=matrix, loads=None, **kwargs)

    def set_distortion_coefficients(self, **kwargs: float) -> None:
        """Set distortion coefficients of camera intrinsics.

        :param kwargs: Keyword arguments containing p1, p2, ..., k1, k2, ...
        """
        self._distortion_coefficients = DistortionCoefficients(loads=None, **kwargs)

    def project(self, point: Sequence[float], is_fisheye: bool = False) -> Vector2D:
        """Project a point to the pixel coordinate,
           before that distort the point if ditortion coefficients are provided.

        :param point: A Sequence containing coordinates of the point to be projected
        :return: The coordinate on the pixel plane where the point is projected to
        """
        if self._distortion_coefficients:
            point = self._distortion_coefficients.distort(point, is_fisheye)
        return self._camera_matrix.project(point)



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines some exceptions about TensorBay."""

from requests import RequestException
from requests.models import Response

from .log import dump_request_and_response


class GASException(Exception):
    """This is the parent class to the following specified error classes."""


class GASRequestError(GASException):
    """Through this class which inherits from GASException,
    we can raise GASRequestError specifically.

    :param error: RequestException message
    """

    def __init__(self, error: RequestException) -> None:
        super().__init__()
        self._error = error

    def __str__(self) -> str:
        return str(self._error)


class GASResponseException(GASException):
    """Through this class which inherits from GASException,
    we can raise GASResponseException specifically.
    This is the parent class to response-related errors.

    :param response: The response of the request
    """

    def __init__(self, response: Response) -> None:
        super().__init__()
        self._response = response
        self.status_code = response.status_code

    def __str__(self) -> str:
        return dump_request_and_response(self._response)


class GASResponseError(GASResponseException):
    """Through this class which inherits from GASError,
    we can raise GASResponseError specifically.

    :param response: The response of the request
    """


class GASOSSError(GASResponseException):
    """Exception for OSS response error

    :param response: The response of the request
    """

    def __init__(self, response: Response) -> None:
        super().__init__(response)
        response_json = response.json()
        self.status = response_json["status"]


class GASTensorBayError(GASResponseException):
    """Exception for graviti TensorBay response error

    :param response: The response of the request
    """

    def __init__(self, response: Response) -> None:
        super().__init__(response)
        response_json = response.json()
        self.code = response_json["code"]
        self.message = response_json["message"]
        self.success = response_json["success"]


class GASDatasetError(GASException):
    """Through this class which inherits from GASException,
    we can raise GASDatasetError specifically.

    :param dataset_name: The name of the missing dataset
    """

    def __init__(self, dataset_name: str) -> None:
        super().__init__()
        self._dataset_name = dataset_name

    def __str__(self) -> str:
        return f"Dataset '{self._dataset_name}' does not exist"


class GASDatasetTypeError(GASException):
    """Through this class which inherits from GASException,
    we can raise GASDatasetTypeError specifically.

    :param dataset_name: The name of the dataset whose requested type is wrong
    :param is_fusion: whether the dataset is a fusion dataset
    """

    def __init__(self, dataset_name: str, is_fusion: bool) -> None:
        super().__init__()
        self._dataset_name = dataset_name
        self._is_fusion = is_fusion

    def __str__(self) -> str:
        return (
            f"Dataset '{self._dataset_name}' is {'' if self._is_fusion else 'not '}a fusion dataset"
        )


class GASLabelsetError(GASException):
    """Through this class which inherits from GASException,
    we can raise GASLabelsetError specifically.

    :param labelset_id: The labelset ID of the missing labelset
    """

    def __init__(self, labelset_id: str) -> None:
        super().__init__()
        self._labelset_id = labelset_id

    def __str__(self) -> str:
        return f"Labelset '{self._labelset_id}' does not exist"


class GASLabelsetTypeError(GASException):
    """Through this class which inherits from GASException,
    we can raise GASLabelsetTypeError specifically.

    :param labelset_id: The ID of the labelset whose requested type is wrong
    :param is_fusion: whether the labelset is a fusion labelset
    """

    def __init__(self, labelset_id: str, is_fusion: bool) -> None:
        super().__init__()
        self._labelset_id = labelset_id
        self._is_fusion = is_fusion

    def __str__(self) -> str:
        return (
            f"Labelset '{self._labelset_id}' is "
            f"{'' if self._is_fusion else 'not '}a fusion labelset"
        )


class GASSegmentError(GASException):
    """Through this class which inherits from GASException,
    we can raise GASSegmentError specifically.

    :param segment_name: The name of the missing segment_name
    """

    def __init__(self, segment_name: str) -> None:
        super().__init__()
        self._segment_name = segment_name

    def __str__(self) -> str:
        return f"Segment '{self._segment_name}' does not exist"


class GASPathError(GASException):
    """Throw this error when remote path does not follow linux style.

    :param remote_path: the invalid remote path
    """

    def __init__(self, remote_path: str) -> None:
        super().__init__()
        self._remote_path = remote_path

    def __str__(self) -> str:
        return f'Invalid path: "{self._remote_path}"\nRemote path should follow linux style.'


class GASFrameError(GASException):
    """Throw this error when the uploading frame has no timestamp and no frame_index """

    def __str__(self) -> str:
        return "Either data.timestamp or frame_index is required to sort frame in TensorBay"



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Segment and FusionSegment."""

from typing import Any, Dict, Iterable, List, MutableSequence, Optional, Union, overload

from ..sensor import Sensor, SensorType
from ..utility import NameClass, NameSortedDict
from .data import Data
from .frame import Frame


class Segment(NameClass, MutableSequence[Data]):
    """This class defines the concept of segment,
    which represents a list of Data.

    :param loads: {
        "name": <str>
        "description": <str>
        "data": [
            data_dict{...},
            data_dict{...},
            ...
            ...
        ]
    }
    """

    def __init__(
        self,
        name: str = "",
        *,
        loads: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._data: List[Data] = []

        if loads:
            super().__init__(loads=loads)
            for data_dict in loads["data"]:
                self._data.append(Data(loads=data_dict))
            return

        super().__init__(name)

    def dumps(self) -> Dict[str, Any]:
        """Dump a Segment into a list.

        :return: A dictionary contains name and annotations
        """

        segment_dict: Dict[str, Any] = super().dumps()
        segment_dict["data"] = [data.dumps() for data in self._data]

        return segment_dict

    def insert(self, index: int, value: Data) -> None:
        """Insert Data before index."""
        return self._data.insert(index, value)

    @overload
    def __getitem__(self, index: int) -> Data:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[Data]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[List[Data], Data]:
        return self._data.__getitem__(index)

    @overload
    def __setitem__(self, index: int, value: Data) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[Data]) -> None:
        ...

    def __setitem__(self, index: Union[int, slice], value: Union[Data, Iterable[Data]]) -> None:
        # https://github.com/python/mypy/issues/7858
        self._data.__setitem__(index, value)  # type: ignore[index, assignment]

    def __delitem__(self, index: Union[int, slice]) -> None:
        self._data.__delitem__(index)

    def __len__(self) -> int:
        return self._data.__len__()


class FusionSegment(NameClass, MutableSequence[Frame]):
    """This class defines the concept of multi-sensor segment,
    which represents a list of Frames.

    :param loads: {
        "name": <str>
        "description": <str>
        "sensors": [
            sensor_dict{...},
            sensor_dict{...},
            ...
            ...
        ]
        "frames": [
            frame_dict{...},
            frame_dict{...},
            ...
            ...
        ]
    }
    """

    def __init__(
        self,
        name: str = "",
        *,
        loads: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._frames: List[Frame] = []
        self._sensors: NameSortedDict[Sensor] = NameSortedDict()

        if loads:
            super().__init__(loads=loads)
            for sensor_dict in loads["sensors"]:
                self._sensors.add(Sensor(loads=sensor_dict))
            for frame_dict in loads["frames"]:
                self._frames.append(Frame(loads=frame_dict))
            return

        super().__init__(name)

    def dumps(self) -> Dict[str, Any]:
        """Dumps the segment into a dictionary.

        :return: A dictonary contains name, sensors and frames
        """

        segment_dict: Dict[str, Any] = super().dumps()
        segment_dict["sensors"] = [sensor.dumps() for sensor in self._sensors.values()]
        segment_dict["frames"] = [frame.dumps() for frame in self._frames]

        return segment_dict

    def insert(self, index: int, value: Frame) -> None:
        """Insert Frame before index."""
        return self._frames.insert(index, value)

    @overload
    def __getitem__(self, index: int) -> Frame:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[Frame]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[List[Frame], Frame]:
        return self._frames.__getitem__(index)

    @overload
    def __setitem__(self, index: int, value: Frame) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[Frame]) -> None:
        ...

    def __setitem__(self, index: Union[int, slice], value: Union[Frame, Iterable[Frame]]) -> None:
        self._frames.__setitem__(index, value)  # type: ignore[index, assignment]

    def __delitem__(self, index: Union[int, slice]) -> None:
        self._frames.__delitem__(index)

    def __len__(self) -> int:
        return self._frames.__len__()

    def add_sensor(self, sensor: Sensor) -> None:
        """Add sensor.

        :param sensor: sensor to add
        """

        self._sensors.add(sensor)

    def get_sensor(self, sensor_name: str) -> Sensor:
        """Return sensor info corresponding to input sensor name."""

        return self._sensors[sensor_name]

    def get_sensors(self, sensor_type: Optional[SensorType] = None) -> NameSortedDict[Sensor]:
        """Return all sensors with given sensor type.

        :param sensor_type: sensor type
        :return: a dict of sensors with given sensor type
        """

        if not sensor_type:
            return self._sensors

        data: NameSortedDict[Sensor] = NameSortedDict()
        for sensor in self._sensors.values():
            if sensor.enum == sensor_type:
                data.add(sensor)
        return data

    def get_data_list(self, sensor: Union[str, Sensor]) -> List[Data]:
        """Return all Datas corresponding to given sensor.

        :param sensor: a sensor name or a sensor class
        :return: a list of Datas corresponding to given sensor
        """

        sensor_name = sensor.name if isinstance(sensor, Sensor) else sensor

        return [frame[sensor_name] for frame in self._frames]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#
# pylint: disable=invalid-name

"""Dataloader of LISA Traffic Light."""

from .LISATrafficLight import LISATrafficLight

__all__ = ["LISATrafficLight"]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Quaternion."""

import math
import warnings
from typing import Dict, Optional, Sequence, Type, TypeVar, Union, overload

import numpy as np

from .vector import Vector3D

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import quaternion


T = TypeVar("T", bound="Quaternion")  # pylint: disable=invalid-name


class Quaternion:
    """Class to represent a 4-dimensional complex number or quaternion.

    Quaternion objects can be used generically as 4D numbers,
    or as unit quaternions to represent rotations in 3D space.

    Quaternion can be initialized in the following ways:
    1. Init from keyword-only argument: "loads"
        >>> Quaternion(loads={"w": 1, "x": 0, "y": 0, "z": 0})
        >>> Quaternion(1.0, 0.0, 0.0, 0.0)

    2. Init from w, x, y, z components
        >>> Quaternion(1)
        >>> Quaternion(1.0, 0.0, 0.0, 0.0)

        >>> Quaternion(w=1, x=0, y=0, z=0)
        >>> Quaternion(1.0, 0.0, 0.0, 0.0)

        >>> Quaternion(1.0, 0.0, 0.0, 0.0)
        >>> Quaternion(1.0, 0.0, 0.0, 0.0)

        >>> Quaternion([1.0, 0.0, 0.0, 0.0])
        >>> Quaternion(1.0, 0.0, 0.0, 0.0)

    3. Init from a 3*3 or 3*4 or 4*4 sequence/numpy array
        >>> Quaternion([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        >>> Quaternion(1.0, 0.0, 0.0, 0.0)

    4. Init from None
        >>> Quaternion(None)
        >>> Quaternion(1.0, 0.0, 0.0, 0.0)

    5. Init from rotation vector
        >>> Quaternion(rotation_vector=[0, 0, 0])
        >>> Quaternion(1.0, 0.0, 0.0, 0.0)

    6. Init from axis angle
        >>> Quaternion(axis=[1, 0, 0], radians=1.57)
        >>> Quaternion(0.7073882691671998, 0.706825181105366, 0.0, 0.0)

        >>> Quaternion(axis=[1, 0, 0], degrees=90)
        >>> Quaternion(0.7073882691671998, 0.706825181105366, 0.0, 0.0)

    7. Init from spherical coordinate
        >>> Quaternion(spherical_coords=[1.57, 0.0])
        >>> Quaternion(0.7073882691671998, -0.0, 0.706825181105366, 0.0)

    8. Init from euler angle
        >>> Quaternion(euler_angle=[0, 1.57, 0])
        >>> Quaternion(0.7073882691671998, -0.0, 0.706825181105366, 0.0)

    :param args: Coordinates of the Quaternion
    :param loads: A dicitionary containing coordinates of a Quaternion
    :param kwargs: keyword-only argument to the Quaternion
    :raises TypeError: When the shape of the input matrix args is not 3x3 or 3x4 or 4x4
    """

    ArgsType = Union[
        None,
        Sequence[float],
        Sequence[Sequence[float]],
        np.ndarray,
        quaternion.quaternion,
        "Quaternion",
    ]
    KwargsType = Union[None, float, Sequence[float], np.ndarray]

    def __init__(
        self,
        *args: Union[ArgsType, float],
        loads: Optional[Dict[str, float]] = None,
        **kwargs: KwargsType,
    ) -> None:
        if loads:
            self._data = self._quaternion_from_wxyz(loads)
            return

        if kwargs:
            self._data = self._quaternion_from_kwargs(kwargs)
            if self._data:
                return

        arg = args[0] if len(args) == 1 else args
        if isinstance(arg, Sequence) and arg:  # pylint: disable=W1116
            arg = np.array(arg, dtype=np.float64)

        if isinstance(arg, np.ndarray):
            if arg.shape == (3, 3) or arg.shape == (3, 4) or arg.shape == (4, 4):
                self._data = quaternion.from_rotation_matrix(arg[:3, :3])
                return

            if len(arg.shape) > 1:
                raise TypeError(
                    "Only support 3x3, 3x4 and 4x4 matrix input"
                    f"to initialize {self.__class__.__name__}"
                )

            self._data = quaternion.quaternion(*arg)
            return

        if not arg:
            self._data = quaternion.quaternion(1, 0, 0, 0)
            return

        if isinstance(arg, Quaternion):
            self._data = quaternion.quaternion(arg._data)
            return

        self._data = quaternion.quaternion(arg)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._data.w}, "
            f"{self._data.x}, "
            f"{self._data.y}, "
            f"{self._data.z})"
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Quaternion):
            return self._data.__eq__(other._data)  # type: ignore[no-any-return]
        return False

    def __neg__(self: T) -> T:
        return self._create(self._data.__neg__())

    def __add__(self: T, other: object) -> T:
        if not isinstance(other, Quaternion):
            return NotImplemented

        return self._create(self._data.__add__(other._data))

    def __sub__(self: T, other: object) -> T:
        if not isinstance(other, Quaternion):
            return NotImplemented

        return self._create(self._data.__sub__(other._data))

    @overload
    def __mul__(self: T, other: T) -> T:
        ...

    @overload
    def __mul__(self: T, other: Sequence[float]) -> Vector3D:
        ...

    @overload
    def __mul__(self: T, other: np.ndarray) -> Vector3D:
        ...

    # mypy errors to be figured out
    # when swapping overload 2 and 3
    # 1) Overloaded function signature 3 will never be matched:
    #    signature 2's parameter type(s) are the same or broader
    # when use Union to describe overload 2 and 3 in one
    # 2) Overloaded function signatures 1 and 2 overlap with incompatible return types

    def __mul__(self: T, other: Union[T, np.ndarray, Sequence[float]]) -> Union[T, Vector3D]:
        if isinstance(other, Quaternion):
            return self._create(self._data.__mul__(other._data))

        if isinstance(other, (Sequence, np.ndarray)):  # pylint: disable=W1116
            return self.rotate(other)

        return NotImplemented

    @classmethod
    def _create(cls: Type[T], data: quaternion.quaternion) -> T:
        obj: T = object.__new__(cls)
        obj._data = data  # pylint: disable=protected-access
        return obj

    @property
    def w(self) -> float:  # pylint: disable=invalid-name
        """Get the w component of the quaternion."""
        return self._data.w  # type: ignore[no-any-return]

    @property
    def x(self) -> float:  # pylint: disable=invalid-name
        """Get the x component of the quaternion."""
        return self._data.x  # type: ignore[no-any-return]

    @property
    def y(self) -> float:  # pylint: disable=invalid-name
        """Get the y component of the quaternion."""
        return self._data.y  # type: ignore[no-any-return]

    @property
    def z(self) -> float:  # pylint: disable=invalid-name
        """Get the z component of the quaternion."""
        return self._data.z  # type: ignore[no-any-return]

    @property
    def radians(self) -> float:
        """Get the angle of the quaternion in radians."""
        return self._data.angle()  # type: ignore[no-any-return]

    @property
    def degrees(self) -> float:
        """Get the angle of the quaternion in degrees."""
        return math.degrees(self._data.angle())

    def as_matrix(self) -> np.ndarray:
        """Return the quaternion as a rotation matrix."""
        return quaternion.as_rotation_matrix(self._data)

    def inverse(self: T) -> T:
        """Get the inverse of this quaternion.
        :return: Inverse quaternion
        """
        return self._create(self._data.inverse())

    def rotate(self, vector: Union[Sequence[float], np.ndarray]) -> Vector3D:
        """Rotate the input vector using this quaternion.
        :return: Rotated vector
        """
        rotated_vector = Vector3D(*quaternion.rotate_vectors(self._data, vector))
        return rotated_vector

    def dumps(self) -> Dict[str, float]:
        """Dumps the Quaternion as a dictionary.

        :return: A dictionary containing Quaternion information
        """
        return {"w": self._data.w, "x": self._data.x, "y": self._data.y, "z": self._data.z}

    @staticmethod
    def _quaternion_from_kwargs(kwargs: Dict[str, KwargsType]) -> Optional[quaternion.quaternion]:
        if "rotation_vector" in kwargs:
            return quaternion.from_rotation_vector(kwargs["rotation_vector"])

        if "spherical_coords" in kwargs:
            return quaternion.from_spherical_coords(kwargs["spherical_coords"])

        if "euler_angle" in kwargs:
            return quaternion.from_euler_angles(kwargs["euler_angle"])

        if "axis" in kwargs:
            if "degrees" in kwargs:
                angle = math.radians(kwargs["degrees"])  # type: ignore[arg-type]
            else:
                angle = kwargs["radians"]  # type: ignore[assignment]
            norm = np.linalg.norm(kwargs["axis"])
            return quaternion.from_rotation_vector(
                [x * angle / norm for x in kwargs["axis"]]  # type: ignore[union-attr]
            )

        if "w" in kwargs or "x" in kwargs or "y" in kwargs or "z" in kwargs:
            return Quaternion._quaternion_from_wxyz(kwargs)  # type: ignore[arg-type]

        return None

    @staticmethod
    def _quaternion_from_wxyz(kwargs: Dict[str, float]) -> quaternion.quaternion:
        return quaternion.quaternion(kwargs["w"], kwargs["x"], kwargs["y"], kwargs["z"])



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#
# pylint: disable=invalid-name

"""This file handles with the LISATrafficLight dataset"""

import csv
import os
import re
from glob import glob

from ...dataset import Data, Dataset, Segment
from ...label import Classification, LabeledBox2D, LabelType

DATASET_NAME = "LISATrafficLight"

SUPERCATEGORY_INDEX = {
    "frameAnnotationsBOX.csv": "BOX",
    "frameAnnotationsBULB.csv": "BULB",
}


def LISATrafficLight(path: str) -> Dataset:
    """LISA traffic light open dataset dataloader

    :param path: Path to LISA traffic light dataset
    The file structure should be like:
    <path>
        Annotations/Annotations/
            daySequence1/
            daySequence2/
            dayTrain/
                dayClip1/
                dayClip10/
                ...
                dayClip9/
            nightSequence1/
            nightSequence2/
            nightTrain/
                nightClip1/
                nightClip2/
                ...
                nightClip5/
        daySequence1/daySequence1/
        daySequence2/daySequence2/
        dayTrain/dayTrain/
            dayClip1/
            dayClip10/
            ...
            dayClip9/
        nightSequence1/nightSequence1/
        nightSequence2/nightSequence2/
        nightTrain/nightTrain/
            nightClip1/
            nightClip2/
            ...
            nightClip5/

    :return: load `Dataset` object
    """
    root_path = os.path.abspath(os.path.expanduser(path))
    annotation_path = os.path.join(root_path, "Annotations", "Annotations")

    dataset = Dataset(DATASET_NAME, is_continuous=True)
    dataset.load_label_tables(os.path.join(os.path.dirname(__file__), "labeltables.json"))

    csv_paths = glob(os.path.join(annotation_path, "**", "*.csv"), recursive=True)
    if len(csv_paths) == 0:
        raise TypeError(f"No '.csv' files found in '{path}'")
    csv_paths.sort()

    for box_csv_path, bulb_csv_path in zip(csv_paths[0::2], csv_paths[1::2]):
        segment = dataset.create_segment(_get_segment_name(box_csv_path))

        prefix = _get_path_prefix(annotation_path, box_csv_path)
        classification = _get_classification(prefix)

        filedir = os.path.join(root_path, prefix)
        filenames = glob(os.path.join(filedir, "*.jpg"))
        if len(filenames) == 0:
            raise TypeError(f"No '.jpg' files found in '{filedir}'")
        filenames.sort()
        # Check the frame_number from filename: "daySequence1--00345.jpg"
        if _get_frame_number(filenames[-1]) + 1 != len(filenames):
            raise TypeError(f"Discontinuous frame number in '{filedir}'")

        for filename in filenames:
            data = Data(filename)
            data.register_label(LabelType.BOX2D)
            if classification:
                data.append_label(Classification(classification))
            segment.append(data)

        _add_labels(segment, box_csv_path)
        _add_labels(segment, bulb_csv_path)

    return dataset


def _get_frame_number(filename: str) -> int:
    return int(filename[-9:-4])


def _get_path_prefix(annotation_path: str, csv_path: str) -> str:
    relpath = os.path.relpath(csv_path, annotation_path)
    splits = relpath.split(os.sep)[:-1]
    return os.path.join(splits[0], *splits, "frames")


def _get_segment_name(csv_path: str) -> str:
    with open(csv_path, "r") as file:
        reader = csv.DictReader(file, delimiter=";")
        first_filename = next(reader)["Filename"]
        header = first_filename[: first_filename.index("-")]
        return re.sub(r"\d+", lambda match: f"{int(match.group(0)):02}", header, 1)


def _get_supercategory(csv_path: str) -> str:
    basename = os.path.basename(csv_path)
    return SUPERCATEGORY_INDEX[basename]


def _get_classification(path_prefix: str) -> str:
    if path_prefix.startswith("day"):
        return "day"

    if path_prefix.startswith("night"):
        return "night"

    return ""


def _add_labels(segment: Segment, csv_path: str) -> None:
    supercategory = _get_supercategory(csv_path)

    with open(csv_path, "r") as file:
        reader = csv.DictReader(file, delimiter=";")
        for row in reader:
            frame_number = int(row["Origin frame number"])
            data = segment[frame_number]

            label = LabeledBox2D(
                int(row["Upper left corner X"]),
                int(row["Upper left corner Y"]),
                int(row["Lower right corner X"]),
                int(row["Lower right corner Y"]),
                category=".".join([supercategory, row["Annotation tag"]]),
            )
            data.append_label(label)



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Label, LabelType, Classification, LabeledBox2D, LabeledBox3D,
LabeledPolygon2D and LabeledPolyline2D
"""

from typing import Any, Dict, Optional, Sequence, TypeVar, Union

from ..geometry import Box2D, Box3D, Polygon2D, Polyline2D, Quaternion, Transform3D
from ..utility import TypeClass, TypeEnum

T = TypeVar("T", bound="LabeledBox3D")  # pylint: disable=invalid-name


class LabelType(TypeEnum):
    """this class defines the type of the labels.

    :param label_key: The key string of the json format label annotation
    """

    CLASSIFICATION = "labels_classification"
    BOX2D = "labels_box2D"
    BOX3D = "labels_box3D"
    POLYGON = "labels_polygon"
    POLYLINE = "labels_polyline"


class Label(TypeClass):
    """this class defines the concept of label and some operations on it.

    :param category: Category of the label
    :param attributes: Attributs of the label
    :param instance: Labeled instance
    :param loads: {
        "category": <str>
        "attributes": <Dict>
        "instance": <str>
    }
    """

    def __init__(
        self,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
        loads: Optional[Dict[str, Any]] = None,
    ):
        if loads:
            self.category = loads.get("category", None)
            self.attributes = loads.get("attributes", None)
            self.instance = loads.get("instance", None)
        else:
            self.category = category
            self.attributes = attributes
            self.instance = instance

    def dumps(self) -> Dict[str, Any]:
        """dump a label into a dict."""

        label_dict: Dict[str, Any] = {}
        if self.category:
            label_dict["category"] = self.category
        if self.attributes:
            label_dict["attributes"] = self.attributes
        if self.instance:
            label_dict["instance"] = self.instance

        return label_dict

    def __repr__(self) -> str:
        str_list = [f"{self.__class__.__name__}("]
        if self.category:
            str_list.append(f'  category: "{self.category}",')
        if self.attributes:
            str_list.append(f"  attributes: {self.attributes},")
        if self.instance:
            str_list.append(f'  instance: "{self.instance}",')
        str_list.append(")")

        return "\n".join(str_list)


class Classification(Label, enum=LabelType.CLASSIFICATION):
    """this class defines the concept of classification label.

    :param category: Category of the label
    :param attributes: Attributs of the label
    :param instance: Labeled instance
    :param loads: {
        "category": <str>
        "attributes": <Dict>
        "instance": <str>
    }
    """


class LabeledBox2D(Box2D, Label, enum=LabelType.BOX2D):
    """Contain the definition of LabeledBox2D bounding box and some related operations.

    :param args: Union[None, float, Sequence[float]],
        box = LabeledBox2D()
        box = LabeledBox2D(10, 20, 30, 40)
        box = LabeledBox2D([10, 20, 30, 40])
    :param category: Category of the label
    :param attributes: Attributs of the label
    :param instance: Labeled instance
    :param loads: {
        "box2D": <List[Dict]>
        "category": <str>
        "attributes": <Dict>
        "instance": <str>
    }
    :param x: X coordinate of the top left vertex of the box
    :param y: Y coordinate of the top left vertex of the box
    :param width: Length along the x axis
    :param height: Length along the y axis
    """

    def __init__(
        self,
        *args: Union[None, float, Sequence[float]],
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
        loads: Optional[Dict[str, Any]] = None,
        x: Optional[float] = None,  # pylint: disable=invalid-name
        y: Optional[float] = None,  # pylint: disable=invalid-name
        width: Optional[float] = None,
        height: Optional[float] = None,
    ):
        if loads:
            Box2D.__init__(self, loads=loads["box2D"])
            Label.__init__(self, loads=loads)
        else:
            Box2D.__init__(self, *args, loads=None, x=x, y=y, width=width, height=height)
            Label.__init__(self, category, attributes, instance)

    def dumps(self) -> Dict[str, Any]:  # type: ignore[override]
        labeled_box2d_dict = Label.dumps(self)
        labeled_box2d_dict["box2D"] = Box2D.dumps(self)
        return labeled_box2d_dict


class LabeledBox3D(Box3D, Label, enum=LabelType.BOX3D):
    """Contain the definition of LabeledBox3D bounding box and some related operations.

    :param transform: A Transform3D object or a 4x4 or 3x4 transfrom matrix
    :param translation: Translation in a sequence of [x, y, z]
    :param rotation: Rotation in a sequence of [w, x, y, z] or 3x3 rotation matrix or `Quaternion`
    :param size: Size in a sequence of [x, y, z]
    :param category: Category of the label
    :param attributes: Attributs of the label
    :param instance: Labeled instance
    :param loads: {
        "translation": translation in a sequence of [x, y, z]
        "rotation": rotation in a sequence of [w, x, y, z]
        "size": size in a sequence of [x, y, z]
        "category": <str>
        "attributes": <Dict>
        "instance": <str>
    }
    :param kwargs: Other parameters to initialize rotation of the transform
    """

    def __init__(
        self,
        transform: Transform3D.TransformType = None,
        *,
        translation: Optional[Sequence[float]] = None,
        rotation: Quaternion.ArgsType = None,
        size: Optional[Sequence[float]] = None,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
        loads: Optional[Dict[str, Any]] = None,
        **kwargs: Quaternion.KwargsType,
    ):
        if loads:
            Box3D.__init__(self, loads=loads["box3D"])
            Label.__init__(self, loads=loads)
            return

        Box3D.__init__(
            self,
            transform,
            translation=translation,
            rotation=rotation,
            size=size,
            loads=None,
            **kwargs,
        )
        Label.__init__(self, category, attributes, instance)

    def __rmul__(self: T, other: Transform3D) -> T:
        if isinstance(other, Transform3D):
            labeled_box_3d = Box3D.__rmul__(self, other)
            labeled_box_3d.category = self.category
            labeled_box_3d.attributes = self.attributes
            labeled_box_3d.instance = self.instance
            return labeled_box_3d

        return NotImplemented  # type: ignore[unreachable]

    def dumps(self) -> Dict[str, Any]:
        labeled_box3d_dict = Label.dumps(self)
        labeled_box3d_dict["box3D"] = Box3D.dumps(self)
        return labeled_box3d_dict


class LabeledPolygon2D(Polygon2D, Label, enum=LabelType.POLYGON):
    """this class defines the polygon2D with labels

    :param points: a list of 2D point list
    :param category: Category of the label
    :param attributes: Attributs of the label
    :param instance: Labeled instance
    :param loads: {
        "polygon": [
            { "x": <int>
              "y": <int>
            },
            ...
            ...
        ],
        "category": <str>
        "attributes": {
            "<key>": "<value>" <str>
            ...
            ...
        }
        "instance": <str>
    }
    """

    def __init__(
        self,
        points: Optional[Sequence[Sequence[float]]] = None,
        *,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
        loads: Optional[Dict[str, Any]] = None,
    ):
        if loads:
            Polygon2D.__init__(self, loads=loads["polygon"])
            Label.__init__(self, loads=loads)
        else:
            Polygon2D.__init__(self, points)
            Label.__init__(self, category, attributes, instance)

    def dumps(self) -> Dict[str, Any]:  # type: ignore[override]
        """dump a LabeledPolygon2D into a dict"""

        data = Label.dumps(self)
        data["polygon"] = Polygon2D.dumps(self)

        return data


class LabeledPolyline2D(Polyline2D, Label, enum=LabelType.POLYLINE):
    """this class defines the polyline2D with labels

    :param points: a list of 2D point list
    :param category: Category of the label
    :param attributes: Attributs of the label
    :param instance: Labeled instance
    :param loads: {
        "polyline": [
            { "x": <int>
              "y": <int>
            },
            ...
            ...
        ],
        "category": <str>
        "attributes": {
            "<key>": "<value>" <str>
            ...
            ...
        }
        "instance": <str>
    }
    """

    def __init__(
        self,
        points: Optional[Sequence[Sequence[float]]] = None,
        *,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
        loads: Optional[Dict[str, Any]] = None,
    ):
        if loads:
            Polyline2D.__init__(self, loads=loads["polyline"])
            Label.__init__(self, loads=loads)
        else:
            Polyline2D.__init__(self, points)
            Label.__init__(self, category, attributes, instance)

    def dumps(self) -> Dict[str, Any]:  # type: ignore[override]
        """dump a LabeledPolyline2D into a dict"""

        data = Label.dumps(self)
        data["polyline"] = Polyline2D.dumps(self)

        return data



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

""" utility classes """

from .name import NameClass, NameSortedDict, NameSortedList
from .tbrn import TBRN, TBRNType
from .type import TypeClass, TypeEnum

__all__ = [
    "NameClass",
    "NameSortedDict",
    "NameSortedList",
    "TBRN",
    "TBRNType",
    "TypeClass",
    "TypeEnum",
]



# !/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""SegmentNameWrapper class for manage the path of the segment object."""


from pathlib import PurePosixPath


class SegmentNameWrapper:
    """SegmentPath class to handle segment path from tensorbay,
    provides functions to convert remote_path and object_path.

    :param segment_name: The name of the segment
    """

    _DEFAULT_SEGMENT_NAME = "_default"
    _PATH_PREFIX: str = ".segment/"
    _PATH_SUFFIX: str = "/.segment_end"

    def __init__(self, segment_name: str) -> None:
        if not segment_name:
            self._segment_prefix = ""
            self._post_segment_name = SegmentNameWrapper._DEFAULT_SEGMENT_NAME
        else:
            self._segment_prefix = (
                SegmentNameWrapper._PATH_PREFIX + segment_name + SegmentNameWrapper._PATH_SUFFIX
            )
            self._post_segment_name = segment_name

    def get_object_path(self, remote_path: str, sensor_name: str = "") -> str:
        """Convert remote_path to object_path

        :param remote_path: The remote_path of the data
        :param sensor_name: The name of the of the data sensor
        :return: The object_path convert from input remote_path
        """
        if not self._segment_prefix and not sensor_name:
            return remote_path

        return str(PurePosixPath(self._segment_prefix, sensor_name, remote_path))

    def get_remote_path(self, object_path: str, sensor_name: str = "") -> str:
        """Convert object_path to remote_path

        :param object_path: The object_path of the data
        :param sensor_name: The name of the of the data sensor
        :return: The remote_path convert from input object_path
        """
        if not self._segment_prefix and not sensor_name:
            return object_path

        header = PurePosixPath(self._segment_prefix, sensor_name)
        return object_path[len(str(header)) + 1 :]

    @property
    def post_segment_name(self) -> str:
        """Get to post segment name of current segment

        :return: The post segment name
        """
        return self._post_segment_name

    @staticmethod
    def get_local_segment_name(name: str) -> str:
        """Convert post segment name to local segment name

        :param name: The post segment name
        :return: The local segment name
        """
        return name if name != SegmentNameWrapper._DEFAULT_SEGMENT_NAME else ""



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file define class Client and function read_data_from_file.
"""

import logging
from concurrent.futures import FIRST_EXCEPTION, ThreadPoolExecutor, wait
from typing import Any, Callable, Dict, Iterable, Optional, TypeVar
from urllib.parse import urljoin

import requests

from .exceptions import GASOSSError, GASRequestError, GASResponseError, GASTensorBayError
from .log import dump_request_and_response

logger = logging.getLogger(__name__)


class Client:
    """This is a base class defining the concept of Client,
    which contains functions to send post to content store and labelset store

    :param access_key: user's access key
    :param url: the url of the graviti gas website
    """

    _DEFAULT_URL = "https://gas.graviti.cn/"

    def __init__(self, access_key: str, url: str = "") -> None:
        if not url:
            url = Client._DEFAULT_URL

        self.gateway_url = urljoin(url, "gatewayv2/")
        self.access_key = access_key

        self._content_store = urljoin(self.gateway_url, "content-store/")
        self._label_store = urljoin(self.gateway_url, "label-store/")

    def dataset_post(self, method: str, post_data: Dict[str, Any]) -> Any:
        """Send a POST request to the TensorBay content-store

        :param method: The method of the request
        :param post_data: Json data to send in the body of the request
        :return: Response of the request
        """
        url = urljoin(self._content_store, method)
        return post(url, json_data=post_data, access_key=self.access_key)

    def labelset_post(self, method: str, post_data: Dict[str, Any]) -> Any:
        """Send a POST request to the TensorBay label-store

        :param method: The method of the request
        :param post_data: Json data to send in the body of the request
        :return: Response of the request
        """
        url = urljoin(self._label_store, method)
        return post(url, json_data=post_data, access_key=self.access_key)


T = TypeVar("T")  # pylint: disable=invalid-name


def multithread_upload(
    function: Callable[[T], None],
    arguments: Iterable[T],
    *,
    jobs: int = 1,
) -> None:
    """multithread upload framework

    :param function: The upload function
    :param arguments: The arguments of the upload function
    :param jobs: The number of the max workers in multithread upload
    """
    with ThreadPoolExecutor(jobs) as executor:
        futures = []
        for argument in arguments:
            futures.append(executor.submit(function, argument))

        done, not_done = wait(futures, return_when=FIRST_EXCEPTION)
        for future in not_done:
            future.cancel()
        for future in done:
            future.result()


def post(
    url: str,
    *,
    data: Optional[bytes] = None,
    json_data: Optional[Dict[str, Any]] = None,
    content_type: Optional[str] = None,
    access_key: Optional[str] = None,
) -> Any:
    """Send a POST requests

    :param url: URL for the request
    :param data: bytes data to send in the body of the request
    :param json_data: json data to send in the body of the request
    :param content_type: Content-Type to send in the header of the request
    :param token: X-Token to send in the header of the request
    :raises GASRequestError: When post request failed
    :raises GASResponseError: When response.ok is False
    :raises GASTensorBayError: When response content 'success' is False
    :raises GASOSSError: When response is return by AliyunOSS and content 'status' is not 'OK'
    :return: response of the request
    """
    headers: Dict[str, str] = {}
    if access_key:
        headers["X-Token"] = access_key
    if content_type:
        headers["Content-Type"] = content_type

    try:
        response = requests.post(url, data=data, json=json_data, headers=headers)
    except requests.RequestException as error:
        raise GASRequestError(error)

    return _parser_response(response)


def get(url: str, params: Dict[str, Any]) -> Any:
    """Send a GET requests

    :param url: URL for the request
    :param params: Dictionary to send in the query string for the `Request`
    :raises GASRequestError: When post request failed
    :raises GASResponseError: When response.ok is False
    :raises GASTensorBayError: When response content 'success' is False
    :return: response of the request
    """
    try:
        response = requests.get(url, params=params, headers={"Accept": "application/json"})
    except requests.RequestException as error:
        raise GASRequestError(error)

    return _parser_response(response)


def _parser_response(response: requests.Response) -> Any:
    if response.status_code != 200:
        raise GASResponseError(response)

    content_type = response.headers["Content-Type"]
    if not content_type.startswith("application/json"):
        logger.debug(dump_request_and_response(response))
        return response.content

    result = response.json()

    if response.headers.get("Server", None) == "AliyunOSS":
        if result["status"] != "OK":
            raise GASOSSError(response)
        logger.debug(dump_request_and_response(response))
        return result

    if not result["success"]:
        raise GASTensorBayError(response)

    logger.debug(dump_request_and_response(response))
    return result["data"]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class GAS."""

from typing import Any, Dict, List, Optional, Tuple, Type, Union, overload
from urllib.parse import urljoin

from typing_extensions import Literal

from ..dataset import Dataset, FusionDataset
from ..label import LabelTables
from .dataset import DatasetClient, FusionDatasetClient
from .exceptions import GASDatasetError, GASDatasetTypeError, GASLabelsetError, GASLabelsetTypeError
from .labelset import FusionLabelsetClient, LabelsetClient, LabelsetClientBase
from .requests import Client, get

DatasetClientType = Union[DatasetClient, FusionDatasetClient]
LabelsetClientType = Union[LabelsetClient, FusionLabelsetClient]


class GAS:
    """This is a class defining the concept of TensorBay.
    It mainly defines some operations on datasets.

    :param access_key: user's access key
    :param url: the url of the gas website
    """

    _VERSIONS = {1: "COMMUNITY", 2: "ENTERPRISE"}

    def __init__(self, access_key: str, url: str = "") -> None:
        self._client = Client(access_key, url)

    def get_user_info(self) -> Dict[str, str]:
        """Get the user info corresponding to the AccessKey

        :return: A directory which contains the username and clientTag
        """
        post_data = {"token": self._client.access_key}
        url = urljoin(self._client.gateway_url, "user/api/v3/token/get-user-profile")
        response = get(url, post_data)
        return {
            "username": response["userName"],
            "version": GAS._VERSIONS[response["clientTag"]],
        }

    @overload
    def _create_dataset(
        self,
        name: str,
        is_continuous: bool,
        is_fusion: Literal[False],
    ) -> DatasetClient:
        ...

    @overload
    def _create_dataset(
        self,
        name: str,
        is_continuous: bool,
        is_fusion: Literal[True],
    ) -> FusionDatasetClient:
        ...

    def _create_dataset(
        self,
        name: str,
        is_continuous: bool,
        is_fusion: bool,
    ) -> DatasetClientType:
        post_data = {
            "name": name,
            "contentSetType": int(is_fusion),  # normal dataset: 0, fusion dataset: 1
            "isContinuous": int(is_continuous),
        }
        data = self._client.dataset_post("createContentSet", post_data)
        dataset_id = data["contentSetId"]
        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(name, dataset_id, self._client)

    def create_dataset(self, name: str, is_continuous: bool = False) -> DatasetClient:
        """Create a dataset with the input name,
        and return the client of the created dataset

        :param name: Name of the dataset, unique for a user
        :param is_continuous: Whether the data in dataset are continuous,
            `True` for continuous data, `False` for Discontinuous data
        :return: The client of the created dataset
        """
        return self._create_dataset(name, is_continuous, False)

    def create_fusion_dataset(self, name: str, is_continuous: bool = False) -> FusionDatasetClient:
        """Create a fusion dataset with the input name,
        and return the client of the created fusion dataset

        :param name: Name of the fusion dataset, unique for a user
        :param is_continuous: Whether the data in dataset are continuous,
            `True` for continuous data, `False` for Discontinuous data
        :return: The client of the created fusion dataset
        """
        return self._create_dataset(name, is_continuous, True)

    def _get_dataset_id_and_type(self, name: str) -> Tuple[str, bool]:
        """Get the dataset ID of the dataset with the input name no matter the type of the dataset

        :param name: The name of the requested dataset
        :raises GASDatasetError: When the requested dataset does not exist
        :return: The tuple of dataset ID and type (`True` for fusion dataset)
        """
        if not name:
            raise GASDatasetError(name)

        datasets_info = self._list_datasets_info(name)
        if not datasets_info:
            raise GASDatasetError(name)

        response = datasets_info[0]["contentSetResp"]
        is_fusion = bool(response["contentSetType"])
        return (response["contentSetId"], is_fusion)

    def _get_dataset(self, name: str) -> DatasetClientType:
        """Get the client of the dataset with the input name no matter the type of the dataset

        :param name: The name of the requested dataset
        :raises GASDatasetError: When the requested dataset does not exist
        :return: The client of the request dataset
        """
        dataset_id, is_fusion = self._get_dataset_id_and_type(name)
        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(name, dataset_id, self._client)

    def get_dataset(self, name: str) -> DatasetClient:
        """Get the client of the dataset with the input name

        :param name: The name of the requested dataset
        :raises GASDatasetError: When the requested dataset does not exist
        :raises GASDatasetTypeError: When requested dataset is a fusion dataset
        :return: The client of the request dataset
        """
        client = self._get_dataset(name)
        if not isinstance(client, DatasetClient):
            raise GASDatasetTypeError(name, True)

        return client

    def get_fusion_dataset(self, name: str) -> FusionDatasetClient:
        """Get the client of the fusion dataset with the input name

        :param name: The name of the requested fusion dataset
        :raises GASDatasetError: When the requested dataset does not exist
        :raises GASDatasetTypeError: When requested dataset is not a fusion dataset
        :return: The client of the request fusion dataset
        """
        client = self._get_dataset(name)
        if not isinstance(client, FusionDatasetClient):
            raise GASDatasetTypeError(name, False)

        return client

    def get_or_create_dataset(self, name: str, is_continuous: bool = False) -> DatasetClient:
        """Get a dataset if 'name' exists. Create one otherwise.

        :param name: The name of a dataset
        :param is_continuous: Whether the data in dataset are continuous,
            `True` for continuous data, `False` for Discontinuous data
        :raises GASDatasetTypeError: When requested dataset is a fusion dataset
        :return: created dataset
        """
        try:
            return self.get_dataset(name)
        except GASDatasetError:
            return self.create_dataset(name, is_continuous)

    def get_or_create_fusion_dataset(
        self,
        name: str,
        is_continuous: bool = False,
    ) -> FusionDatasetClient:
        """Get a dataset if 'name' exists. Create one otherwise.

        :param name: The name of a dataset
        :param is_continuous: Whether the data in dataset are continuous,
            `True` for continuous data, `False` for Discontinuous data
        :raises GASDatasetTypeError: When requested dataset is not a fusion dataset
        :return: created dataset
        """
        try:
            return self.get_fusion_dataset(name)
        except GASDatasetError:
            return self.create_fusion_dataset(name, is_continuous)

    def _create_labelset(
        self,
        dataset_id: str,
        label_tables: LabelTables,
        remote_paths: Optional[List[str]] = None,
    ) -> str:
        """Create a labelset and return the labelset ID.

        :param label_tables: A LabelTables covers all labels of the labelset
        :param remote_paths: A list of remote paths
        :return: The labelset ID of created labelset
        """
        post_data = {
            "contentSetId": dataset_id,
            "type": LabelsetClientBase.TYPE_GROUND_TRUTH,
            "version": "v1.0.2",
        }
        metadata = LabelsetClientBase.create_metadata(label_tables)
        if metadata:
            post_data["meta"] = metadata
        if remote_paths:
            post_data["objectPaths"] = remote_paths

        data = self._client.labelset_post("createLabelSet", post_data)
        labelset_id = data["labelSetId"]
        return labelset_id  # type: ignore[no-any-return]

    def create_labelset(
        self,
        dataset_name: str,
        label_tables: LabelTables,
        remote_paths: Optional[List[str]] = None,
    ) -> LabelsetClient:
        """Create a labelset.

        :param label_tables: A LabelTables covers all labels of the labelset
        :param remote_paths: A list of remote paths
        :return: The client of the created labelset
        """
        dataset_id, is_fusion = self._get_dataset_id_and_type(dataset_name)
        if is_fusion:
            raise GASDatasetTypeError(dataset_name, True)

        labelset_id = self._create_labelset(dataset_id, label_tables, remote_paths)
        return LabelsetClient(labelset_id, dataset_name, self._client)

    def create_fusion_labelset(
        self,
        dataset_name: str,
        label_tables: LabelTables,
        remote_paths: Optional[List[str]] = None,
    ) -> FusionLabelsetClient:
        """Create a labelset.

        :param label_tables: A LabelTables covers all labels of the labelset
        :param remote_paths: A list of remote paths
        :return: The client of the created fusion labelset
        """
        dataset_id, is_fusion = self._get_dataset_id_and_type(dataset_name)
        if not is_fusion:
            raise GASDatasetTypeError(dataset_name, False)

        labelset_id = self._create_labelset(dataset_id, label_tables, remote_paths)
        return FusionLabelsetClient(labelset_id, dataset_name, self._client)

    def _get_labelset_type_and_dataset_name(self, labelset_id: str) -> Tuple[bool, str]:
        if not labelset_id:
            raise GASLabelsetError(labelset_id)

        post_data = {
            "id": labelset_id,
            "projection": {"contentSetType": 1, "contentSetName": 1},
        }

        summaries = self._client.labelset_post("listLabelSetSummaries", post_data)[
            "labelSetSummaries"
        ]
        if not summaries:
            raise GASLabelsetError(labelset_id)

        return (bool(summaries[0]["contentSetType"]), summaries[0]["contentSetName"])

    def get_labelset(self, labelset_id: str) -> LabelsetClient:
        """Get the client of the labelset with the input labelset_id

        :param labelset_id: The labelset ID of the requested labelset
        :raises GASLabelsetError: When the requested labelset does not exist
        :return: The client of the requested labelset
        """
        is_fusion, dataset_name = self._get_labelset_type_and_dataset_name(labelset_id)
        if is_fusion:
            raise GASLabelsetTypeError(labelset_id, True)

        return LabelsetClient(labelset_id, dataset_name, self._client)

    def get_fusion_labelset(self, labelset_id: str) -> FusionLabelsetClient:
        """Get the client of the fusion labelset with the input labelset_id

        :param labelset_id: The labelset ID of the requested fusion labelset
        :raises GASLabelsetError: When the requested fusion labelset does not exist
        :return: The client of the requested fusion labelset
        """
        is_fusion, dataset_name = self._get_labelset_type_and_dataset_name(labelset_id)
        if not is_fusion:
            raise GASLabelsetTypeError(labelset_id, False)

        return FusionLabelsetClient(labelset_id, dataset_name, self._client)

    def list_labelsets(self, dataset_name: str) -> List[str]:
        """List ids of all labelsets of the dataset.

        :param dataset_name: list the labelset with input dataset name. If `None`, list all
        :return: A list of labelsets ids
        """
        dataset_id, _ = self._get_dataset_id_and_type(dataset_name)
        post_data = {
            "contentSetId": dataset_id,
            "projection": {"id": 1},
            "ignoreContentSetInfo": True,
            "pageSize": -1,
        }

        summaries = self._client.labelset_post("listLabelSetSummaries", post_data)[
            "labelSetSummaries"
        ]
        return [summary["id"] for summary in summaries]

    def delete_labelset(self, labelset_id: str) -> None:
        """Delete a labelset according to a labelset id.

        :param labelset_id: The id of the labelset to be deleted
        """
        post_data = {"labelSetId": labelset_id}
        self._client.labelset_post("deleteLabelSet", post_data)

    @overload
    def upload_dataset_object(
        self,
        dataset: Dataset,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> DatasetClient:
        ...

    @overload
    def upload_dataset_object(
        self,
        dataset: FusionDataset,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> FusionDatasetClient:
        ...

    @overload
    def upload_dataset_object(
        self,
        dataset: Union[Dataset, FusionDataset],
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> DatasetClientType:
        ...

    def upload_dataset_object(
        self,
        dataset: Union[Dataset, FusionDataset],
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> DatasetClientType:
        """Upload a `Dataset` or `FusionDataset` to TensorBay,
        This function will upload all info contains in the `Dataset` or `FusionDataset` object,
        which includes:
        - Create a dataset using the name and type of input `Dataset` or `FusionDataset`,
        - Upload all `Segment` or `FusionSegment` in the dataset to TensorBay

        :param dataset: The `Dataset` or `FusionDataset` object needs to be uploaded.
        :param jobs: The number of the max workers in multithread upload
        :param skip_uploaded_files: Set it to `True` to skip the uploaded files
        :return: The `DatasetClient` or `FusionDatasetClient` used for uploading the dataset
        """
        dataset_client: DatasetClientType

        if isinstance(dataset, FusionDataset):
            dataset_client = self.get_or_create_fusion_dataset(dataset.name, dataset.is_continuous)
        else:
            dataset_client = self.get_or_create_dataset(dataset.name, dataset.is_continuous)

        for segment in dataset:
            dataset_client.upload_segment_object(
                segment,  # type: ignore[arg-type]
                jobs=jobs,
                skip_uploaded_files=skip_uploaded_files,
            )

        return dataset_client

    def list_datasets(self) -> List[str]:
        """List names of all datasets.

        :return: A list of names of all datasets
        """
        datasets_info = self._list_datasets_info()
        dataset_names: List[str] = []
        for dataset_info in datasets_info:
            dataset_name = dataset_info["contentSetResp"]["name"]
            dataset_names.append(dataset_name)
        return dataset_names

    def delete_dataset(self, name: str) -> None:
        """Delete a dataset according to its name.

        :param name: The name of the dataset to delete
        :raises GASDatasetError: When the requested dataset does not exist
        """
        dataset_id, _ = self._get_dataset_id_and_type(name)
        post_data = {"contentSetId": dataset_id, "name": name}
        self._client.dataset_post("deleteContentSets", post_data)

    def _list_datasets_info(self, name: Optional[str] = None) -> List[Any]:
        """List info of all datasets.

        :param name: dataset name to list its info. If None, list info of all datasets
        :return: A list of dicts containing dataset info. If name does not exist,
            return an empty list.
        """
        post_data = {"name": name, "pageSize": -1}
        data = self._client.dataset_post("listContentSets", post_data)
        return data["contentSets"]  # type: ignore[no-any-return]

    @overload
    def upload_labelset_object(
        self,
        dataset: Dataset,
        labelset_id: Optional[str] = None,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> LabelsetClient:
        ...

    @overload
    def upload_labelset_object(
        self,
        dataset: FusionDataset,
        labelset_id: Optional[str] = None,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> FusionLabelsetClient:
        ...

    @overload
    def upload_labelset_object(
        self,
        dataset: Union[Dataset, FusionDataset],
        labelset_id: Optional[str] = None,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> LabelsetClientType:
        ...

    def upload_labelset_object(
        self,
        dataset: Union[Dataset, FusionDataset],
        labelset_id: Optional[str] = None,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> LabelsetClientType:
        """Upload the label in `Dataset` or `FusionDataset` to TensorBay,
        This function will upload all label info contains in the `Dataset` or `FusionDataset`,
        which includes:
        - Create a labelset using the name and type of input `Dataset` or `FusionDataset`,
        - Upload all `Segment` or `FusionSegment` in the dataset to TensorBay

        :param dataset: The `Dataset` or `FusionDataset` object needs to be uploaded.
        :param lableset_id: The target labelset_id, if empty, new labelset will be created
        :param jobs: The number of the max workers in multithread upload
        :param skip_uploaded_files: Set it to `True` to skip the uploaded files
        :return: The `LabelsetClient` or `FusionLabelsetClient` used for uploading the label
        """
        labelset_client: LabelsetClientType

        if isinstance(dataset, FusionDataset):
            if labelset_id:
                labelset_client = self.get_fusion_labelset(labelset_id)
            else:
                labelset_client = self.create_fusion_labelset(dataset.name, dataset.label_tables)
        else:
            if labelset_id:
                labelset_client = self.get_labelset(labelset_id)
            else:
                labelset_client = self.create_labelset(dataset.name, dataset.label_tables)

        for segment in dataset:
            labelset_client.upload_segment_object(
                segment,  # type: ignore[arg-type]
                jobs=jobs,
                skip_uploaded_files=skip_uploaded_files,
            )

        return labelset_client



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class DatasetBase, Dataset and FusionDataset."""

import json
from typing import Any, Dict, Sequence, Set, TypeVar, Union, overload

from ..label import LabelTable, LabelTables, LabelType
from ..utility import NameClass, NameSortedList
from .segment import FusionSegment, Segment

T = TypeVar("T", FusionSegment, Segment)  # pylint: disable=invalid-name


class DatasetBase(NameClass, Sequence[T]):
    """This class defines the concept of DatasetBase,
    which represents a whole dataset contains several segments.

    :param name: Name of the dataset
    :param is_continuous: Whether the data in dataset is continuous
    """

    def __init__(self, name: str, is_continuous: bool = False) -> None:
        super().__init__(name)
        self._segments: NameSortedList[T] = NameSortedList()
        self._label_tables: LabelTables = LabelTables()
        self._is_continuous = is_continuous

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[T]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[Sequence[T], T]:
        return self._segments.__getitem__(index)

    def __len__(self) -> int:
        return self._segments.__len__()

    def __str__(self) -> str:
        return f"{super()} {self._segments}"

    @property
    def is_continuous(self) -> bool:
        """Check whether the data in dataset is continuous

        :return: Return `True` if the data is continuous, otherwise return `False`
        """
        return self._is_continuous

    @property
    def label_tables(self) -> LabelTables:
        """Return label tables

        :return: label tables
        """
        return self._label_tables

    def load_label_tables(self, loads: Union[Dict[str, Dict[str, Any]], str]) -> None:
        """Load label tables from json object.

        :param loads: LabelTables dict or the name of the file which contains the LabelTables dict
        """
        if isinstance(loads, str):
            with open(loads) as file:
                label_tables_dict = json.load(file)
        else:
            label_tables_dict = loads

        self._label_tables = LabelTables(loads=label_tables_dict)

    def create_label_table(self, label_type: LabelType, is_tracking: bool = False) -> LabelTable:
        """Create a new label table with given label type and add it to label tables.

        :param label_type: the label type of the label table to create
        :param is_tracking: indicates whether the label table is tracking related

        :return: the created label table
        """

        return self._label_tables.create_label_table(label_type, is_tracking)

    def get_label_table(self, label_type: LabelType) -> LabelTable:
        """return the label table corresponding to given LabeleType.

        :param label_type: a instance of LabelType
        :return: a label table
        """

        return self._label_tables[label_type]

    def get_label_types(self) -> Set[LabelType]:
        """return a set contains all label types.

        :return: a set of all label types
        """

        return set(self._label_tables.keys())

    def get_segment_by_name(self, name: str) -> T:
        """return the segment corresponding to given name.

        :param name: name of the segment
        :return: the segment which matches the input name
        """

        return self._segments.get_from_name(name)

    def add_segment(self, segment: T) -> None:
        """add segment to segment list.

        :param segment: a segment to be added
        """

        self._segments.add(segment)


class Dataset(DatasetBase[Segment]):
    """This class defines the concept of dataset,
    which contains a list of segments.
    """

    def create_segment(self, segment_name: str = "") -> Segment:
        """create a segment with the given name.

        :param segment_name: The name of the created segment
        :return: The created segment
        """
        segment = Segment(segment_name)
        self._segments.add(segment)
        return segment


class FusionDataset(DatasetBase[FusionSegment]):
    """This class defines the concept of multi-sensor dataset,
    which contains a list of multi-sensor segments.
    """

    def create_segment(self, segment_name: str = "") -> FusionSegment:
        """create a fusion segment with the given name.

        :param segment_name: The name of the created fusion segment
        :return: The created fusion segment
        """
        segment = FusionSegment(segment_name)
        self._segments.add(segment)
        return segment



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class SensorType, Sensor, Lidar, Radar, Camera and FisheyeCamera."""

from typing import Any, Dict, Optional, Sequence, Type, TypeVar, Union

from ..geometry import Quaternion, Transform3D
from ..utility import NameClass, TypeClass, TypeEnum
from .intrinsics import CameraIntrinsics

T = TypeVar("T", bound="Sensor")  # pylint: disable=invalid-name


class SensorType(TypeEnum):
    """this class defines the type of the sensors.

    :param sensor_name: The name string of the json format sensor
    """

    LIDAR = "LIDAR"
    RADAR = "RADAR"
    CAMERA = "CAMERA"
    FISHEYE_CAMERA = "FISHEYE_CAMERA"


class Sensor(NameClass, TypeClass):
    """A class representing sensor including
    name, description, translation and rotation.

    :param name: A string representing the sensor's name
    :loads: A dictionary containing name, description and sensor extrinsics
    :raises
        TypeError: Can't instantiate abstract class Sensor
        TypeError: Name required when not given loads
    """

    def __new__(
        cls: Type[T],
        name: Optional[str] = None,  # pylint: disable=unused-argument
        *,
        loads: Optional[Dict[str, Any]] = None,
    ) -> T:
        obj: T
        if loads:
            obj = object.__new__(SensorType(loads["type"]).type)
            return obj

        if cls is Sensor:
            raise TypeError("Can't instantiate abstract class Sensor")

        obj = object.__new__(cls)
        return obj

    def __init__(
        self, name: Optional[str] = None, *, loads: Optional[Dict[str, Any]] = None
    ) -> None:
        if loads:
            super().__init__(loads=loads)
            if "extrinsics" not in loads:
                self._extrinsics = None
            else:
                self._extrinsics = Transform3D(loads=loads["extrinsics"])
            return

        super().__init__(name)
        self._extrinsics = None

    def dumps(self) -> Dict[str, Any]:
        """Dumps the sensor as a dictionary.

        :return: A dictionary containing name, description and extrinsics
        """
        data: Dict[str, Any] = super().dumps()
        data["type"] = self.enum.value
        if self._extrinsics:
            data["extrinsics"] = self._extrinsics.dumps()

        return data

    @property
    def extrinsics(self) -> Optional[Transform3D]:
        """Return extrinsics of the sensor.

        :return: Extrinsics of the sensor
        """
        return self._extrinsics

    def set_extrinsics(
        self,
        transform: Transform3D.TransformType = None,
        *,
        translation: Optional[Sequence[float]] = None,
        rotation: Quaternion.ArgsType = None,
        loads: Optional[Dict[str, Dict[str, float]]] = None,
        **kwargs: Quaternion.KwargsType,
    ) -> None:
        """Set the extrinsics of the sensor.

        :param transform: A Transform3D object representing the extrinsics
        :param translation: Translation in a sequence of [x, y, z]
        :param rotation: Rotation in a sequence of [w, x, y, z] or 3x3 rotation matrix or Quaternion
        :param loads: A dictionary containing the translation and the rotation
        :param matrix: A 4x4 transform matrix
        :param kwargs: Other parameters to initialize rotation
        """
        self._extrinsics = Transform3D(
            transform, translation=translation, rotation=rotation, loads=loads, **kwargs
        )

    def set_translation(
        self,
        *args: Union[float, Sequence[float]],
        **kwargs: float,
    ) -> None:
        """Set the translation of the sensor.

        :param args: Coordinates of the translation vector
        :param kwargs: keyword-only argument to set different dimension for translation vector
            sensor.set_translation(x=1, y=2, z=3)
        """
        if not self._extrinsics:
            self._extrinsics = Transform3D()
        self._extrinsics.set_translation(*args, **kwargs)

    def set_rotation(
        self,
        *args: Union[Quaternion.ArgsType, float],
        **kwargs: Quaternion.KwargsType,
    ) -> None:
        """Set the rotation of the sensor.

        :param args: Coordinates of the Quaternion
        :param kwargs: keyword-only argument to the Quaternion
        """
        if not self._extrinsics:
            self._extrinsics = Transform3D()
        self._extrinsics.set_rotation(*args, **kwargs)


class Lidar(Sensor, enum=SensorType.LIDAR):
    """This class defines the concept of lidar."""


class Radar(Sensor, enum=SensorType.RADAR):
    """This class defines the concept of radar."""


class Camera(Sensor, enum=SensorType.CAMERA):
    """A class representing camera including
    name, description, translation , rotation, cameraMatrix and distortionCoefficients.

    :param name: A string representing camera's name
    :param loads: A dictionary containing name, description, extrinsics and intrinsics
    """

    def __init__(
        self,
        name: Optional[str] = None,
        *,
        loads: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(name, loads=loads)
        if not loads or "intrinsics" not in loads:
            self._intrinsics = None
        else:
            self._intrinsics = CameraIntrinsics(loads=loads["intrinsics"])

    def dumps(self) -> Dict[str, Any]:
        """Dumps the camera as a dictionary.

        :return: A dictionary containing name, description, extrinsics and intrinsics
        """
        message = super().dumps()
        if self._intrinsics:
            message["intrinsics"] = self._intrinsics.dumps()
        return message

    @property
    def intrinsics(self) -> Optional[CameraIntrinsics]:
        """Return intrinsics."""

        return self._intrinsics

    def set_camera_matrix(
        self,
        matrix: Optional[Sequence[Sequence[float]]] = None,
        **kwargs: float,
    ) -> None:
        """Set camera matrix."""

        if not self._intrinsics:
            self._intrinsics = CameraIntrinsics(
                matrix, loads=None, _init_distortion=False, **kwargs
            )
            return

        self._intrinsics.set_camera_matrix(matrix, **kwargs)

    def set_distortion_coefficients(self, **kwargs: float) -> None:
        """Set distortion coefficients.
        :raise VauleError: When camera matrix of intrinsics is not set yet.
        """

        if not self._intrinsics:
            raise ValueError("Camera matrix of camera intrinsics must be set first.")
        self._intrinsics.set_distortion_coefficients(**kwargs)


class FisheyeCamera(Camera, enum=SensorType.FISHEYE_CAMERA):
    """This class defines the concept of fisheye camera."""



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

""" geometry classes """

from .box import Box2D, Box3D
from .polygon import Polygon2D
from .polyline import Polyline2D
from .quaternion import Quaternion
from .transform import Transform3D
from .vector import Vector, Vector2D, Vector3D

__all__ = [
    "Box2D",
    "Box3D",
    "Polygon2D",
    "Polyline2D",
    "Quaternion",
    "Transform3D",
    "Vector",
    "Vector2D",
    "Vector3D",
]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

""" dataset classes """

from .checkhealth import checkhealth
from .data import Data
from .dataset import Dataset, FusionDataset
from .frame import Frame
from .segment import FusionSegment, Segment

__all__ = [
    "Data",
    "Dataset",
    "Frame",
    "FusionDataset",
    "FusionSegment",
    "Segment",
    "checkhealth",
]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class DatasetClientBase, DatasetClient, FusionDatasetClient,
DataSegmentClientBase, DataSegmentClient, FusionDataSegmentClient
"""

import base64
import json
import os
import threading
import time
import uuid
from copy import deepcopy
from pathlib import PurePosixPath
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, TypeVar, Union

import filetype
from requests_toolbelt import MultipartEncoder

from ..dataset import Data, Frame, FusionSegment, Segment
from ..sensor import Camera, FisheyeCamera, Sensor, SensorType
from ..utility import TBRN, NameSortedDict
from .exceptions import GASException, GASFrameError, GASPathError, GASSegmentError
from .path import SegmentNameWrapper
from .requests import Client, multithread_upload, post

T = TypeVar("T", bound=Sensor)  # pylint: disable=invalid-name


class DataSegmentClientBase:
    """This class defines the concept of a segment client and some opertions on it.

    :param name: Segment name, unique for a dataset
    :param dataset_name: Dataset name
    :param dataset_id: Dataset id
    :param client: The client used for sending request to TensorBay
    """

    _PERMISSION_CATEGORY: str
    _EXPIRED_IN_SECOND = 240

    def __init__(self, name: str, dataset_name: str, dataset_id: str, client: Client) -> None:
        self._name = name
        self._dataset_name = dataset_name
        self._dataset_id = dataset_id
        self._client = client
        self._permission = {"expireAt": 0}
        self._permission_lock = threading.Lock()
        self._name_wrapper = SegmentNameWrapper(name)

    def upload_description(self, description: Optional[str] = None) -> None:
        """Upload description of the segment client.

        :param description: Description of the segment client to upload
        """
        if not description:
            return
        post_data = {"contentSetId": self._dataset_id, "name": self._name, "desc": description}
        self._client.dataset_post("createOrUpdateSegment", post_data)

    def get_segment_name(self) -> str:
        """Return the name of this segment client."""
        return self._name

    def _get_upload_permission(self) -> Dict[str, Any]:
        with self._permission_lock:
            if int(time.time()) >= self._permission["expireAt"]:
                post_data = {
                    "id": self._dataset_id,
                    "category": self._PERMISSION_CATEGORY,
                    "expiredInSec": self._EXPIRED_IN_SECOND,
                    "segmentName": self._name,
                }
                self._permission = self._client.dataset_post("getPutPermission", post_data)

            return deepcopy(self._permission)

    def _clear_upload_permission(self) -> None:
        with self._permission_lock:
            self._permission = {"expireAt": 0}

    @staticmethod
    def _post_multipart_formdata(
        url: str,
        local_path: str,
        remote_path: str,
        data: Dict[str, Any],
    ) -> None:
        with open(local_path, "rb") as file:
            data["file"] = (remote_path, file, filetype.guess_mime(local_path))
            multipart = MultipartEncoder(data)
            post(url, data=multipart, content_type=multipart.content_type)


class DataSegmentClient(DataSegmentClientBase):
    """DataSegmentClient has only one sensor, supporting upload_data method."""

    _PERMISSION_CATEGORY = "contentSet"

    def upload_data(self, local_path: str, remote_path: str = "") -> None:
        """Upload data with local path to the segment.

        :param local_path: The local path of the data to upload
        :param remote_path: The path to save the data in segment client
        :raises
            GASPathError: when remote_path does not follow linux style
            GASFrameError: when uploading frame has neither timestamp nor frame_index
        """

        if not remote_path:
            remote_path = os.path.basename(local_path)

        if "\\" in remote_path:
            raise GASFrameError()

        permission = self._get_upload_permission()
        post_data = permission["result"]
        post_data["key"] = permission["extra"]["objectPrefix"] + remote_path

        try:
            self._post_multipart_formdata(
                permission["extra"]["host"],
                local_path,
                remote_path,
                post_data,
            )
        except GASException:
            self._clear_upload_permission()
            raise

    def upload_data_object(self, data: Data) -> None:
        """Upload data with local path in Data object to the segment.

        :param data: The data object which represents the local file to upload
        """
        self.upload_data(data.local_path, data.remote_path)

    def delete_data(self, remote_paths: Union[str, List[str]]) -> None:
        """Delete data with remote paths.

        :param remote_path_list: A single path or a list of paths which need to deleted,
        eg: test/os1.png or [test/os1.png]
        """
        if not isinstance(remote_paths, list):
            remote_paths = [remote_paths]

        object_paths = [self._name_wrapper.get_object_path(path) for path in remote_paths]
        post_data = {
            "contentSetId": self._dataset_id,
            "filePaths": object_paths,
        }
        self._client.dataset_post("deleteObjects", post_data)

    def list_data(self) -> List[str]:
        """List all data in a segment client.

        :return: A list of data path
        """
        post_data = {
            "contentSetId": self._dataset_id,
            "segmentName": self._name_wrapper.post_segment_name,
        }
        data = self._client.dataset_post("listObjects", post_data)
        return [self._name_wrapper.get_remote_path(object_path) for object_path in data["objects"]]

    def list_data_objects(self) -> List[Data]:
        """List all `Data` object in a dataset segment.

        :return: A list of `Data` object
        """
        data_list = []
        for remote_path in self.list_data():
            tbrn = TBRN(self._dataset_name, self._name, remote_path=remote_path)
            data_list.append(Data(tbrn))
        return data_list


class FusionDataSegmentClient(DataSegmentClientBase):
    """FusionSegmentClient has multiple sensors,
    supporting upload_sensor and upload_frame method.
    """

    _PERMISSION_CATEGORY = "frame"

    def upload_sensor_object(self, sensor: T) -> None:
        """Upload sensor to the segment client.

        :param sensor: The sensor to upload
        """
        post_data = {
            "name": sensor.name,
            "type": sensor.enum.value,
            "extrinsicParams": sensor.extrinsics.dumps() if sensor.extrinsics else {},
        }

        if isinstance(sensor, (Camera, FisheyeCamera)):
            if sensor.intrinsics:
                intrinsics = sensor.intrinsics.dumps()
                intrinsics["distortionCoefficient"] = intrinsics.pop("distortionCoefficients", {})
                post_data["intrinsicParams"] = intrinsics
            else:
                post_data["intrinsicParams"] = {}

        if sensor.description:
            post_data["desc"] = sensor.description

        post_data["contentSetId"] = self._dataset_id
        post_data["segmentName"] = self._name
        self._client.dataset_post("createOrUpdateSensor", post_data)

    def delete_sensors(self, sensor_names: Union[str, List[str]]) -> None:
        """Delete sensors with a single name or a name list.

        :param sensor_names: A single sensor name or a list of sensor names
        """
        if not isinstance(sensor_names, list):
            sensor_names = [sensor_names]
        post_data = {
            "contentSetId": self._dataset_id,
            "sensorNames": {self._name: sensor_names},
        }
        self._client.dataset_post("deleteSensors", post_data)

    def _list_sensor_summaries(self) -> List[Dict[str, Any]]:
        post_data = {
            "contentSetId": self._dataset_id,
            "segmentName": self._name_wrapper.post_segment_name,
        }
        data = self._client.dataset_post("listSensors", post_data)
        return data["sensors"]  # type: ignore[no-any-return]

    def list_sensors(self) -> List[str]:
        """List all sensor names in a segment client.

        :return: A list of sensor name
        """
        sensor_summaries = self._list_sensor_summaries()
        sensor_names: List[str] = []
        for sensor_info in sensor_summaries:
            sensor_names.append(sensor_info["name"])
        return sensor_names

    def list_sensor_objects(self) -> NameSortedDict[Sensor]:
        """List all sensors in a segment client.

        :return: A NameSortedDict of `Sensor` object
        """
        sensor_summaries = self._list_sensor_summaries()
        sensors: NameSortedDict[Sensor] = NameSortedDict()
        for sensor_info in sensor_summaries:
            SensorClass = SensorType(sensor_info["type"]).type  # pylint: disable=invalid-name
            sensor = SensorClass(sensor_info["name"])
            sensor.description = sensor_info["desc"]
            extrinsics = json.loads(sensor_info["extrinsicParams"])
            if extrinsics:
                sensor.set_extrinsics(loads=extrinsics)
            if isinstance(sensor, Camera):
                intrinsics = json.loads(sensor_info["intrinsicParams"])
                if intrinsics:
                    sensor.set_camera_matrix(**intrinsics["cameraMatrix"])
                    if "distortionCoefficient" in intrinsics:
                        sensor.set_distortion_coefficients(**intrinsics["distortionCoefficient"])
            sensors.add(sensor)

        return sensors

    def upload_frame_object(self, frame: Frame, frame_index: Optional[int] = None) -> None:
        """Upload frame to the segment client.

        :param frame: The frame to upload
        :param frame_index: The frame index, used for TensorBay to sort the frame
        :raises
            GASPathError: when remote_path does not follow linux style
        """

        frame_id = str(uuid.uuid4())

        for sensor_name, data in frame.items():
            remote_path = data.remote_path

            if "\\" in remote_path:
                raise GASPathError(remote_path)

            if frame_index is None and data.timestamp is None:
                raise TypeError(
                    "Either 'frame_index' or 'timestamp' is necessary for sorting frames"
                )

            permission = self._get_upload_permission()
            post_data = permission["result"]

            path = str(PurePosixPath(sensor_name, remote_path))
            post_data["key"] = permission["extra"]["objectPrefix"] + path

            data_info: Dict[str, Any] = {
                "sensorName": sensor_name,
                "segmentName": self._name,
                "frameId": frame_id,
                "objectPath": path,
            }
            if data.timestamp is not None:
                data_info["timestamp"] = data.timestamp

            if frame_index is not None:
                data_info["frameIndex"] = frame_index

            post_data["x:incidental"] = base64.urlsafe_b64encode(
                json.dumps(data_info).encode()
            ).decode()

            try:
                self._post_multipart_formdata(
                    permission["extra"]["host"],
                    data.local_path,
                    remote_path,
                    post_data,
                )
            except GASException:
                self._clear_upload_permission()
                raise

    def list_frame_objects(self) -> List[Frame]:
        """List all frames in the segment client.

        :return: A list `Frame` object
        """
        response = self._list_frames()
        frames = []
        for i, response_frame in enumerate(response):
            frame = self._loads_frame_object(i, response_frame)
            frames.append(frame)

        return frames

    def _list_frame_objects_dict(self) -> Dict[int, Tuple[Frame, str]]:
        response = self._list_frames()
        frames: Dict[int, Tuple[Frame, str]] = {}
        for i, response_frame in enumerate(response):
            frame = self._loads_frame_object(i, response_frame)
            frame_index = response_frame[0]["frameIndex"]
            frame_id = response_frame[0]["frameId"]

            frames[frame_index] = (frame, frame_id)

        return frames

    def _list_frames(self, offset: int = 0, page_size: int = -1) -> List[List[Dict[str, Any]]]:
        post_data = {
            "contentSetId": self._dataset_id,
            "segmentName": self._name_wrapper.post_segment_name,
            "offSet": offset,
            "pageSize": page_size,
        }
        data = self._client.dataset_post("listFrames", post_data)
        return data["frames"]  # type: ignore[no-any-return]

    def _loads_frame_object(self, frame_index: int, loads: List[Dict[str, Any]]) -> Frame:
        frame = Frame()
        for data_info in loads:
            sensor_name = data_info["sensorName"]
            remote_path = self._name_wrapper.get_remote_path(data_info["objectPath"], sensor_name)
            timestamp = data_info["timestamp"] if data_info["timestamp"] != -1 else None
            tbrn = TBRN(
                self._dataset_name,
                self._name,
                frame_index,
                sensor_name,
                remote_path=remote_path,
            )
            data = Data(tbrn, timestamp=timestamp)
            frame[sensor_name] = data

        return frame

    def _delete_frames(self, frame_id: Union[str, List[str]]) -> None:
        if not isinstance(frame_id, list):
            frame_id = [frame_id]

        post_data = {"contentSetId": self._dataset_id, "frameIds": frame_id}
        self._client.dataset_post("deleteFrames", post_data)


class DatasetClientBase:
    """This class defines the concept of a dataset and some operations on it.

    :param name: Dataset name
    :param dataset_id: Dataset id
    :param client: The client used for sending request to TensorBay
    """

    _PUBLISH_STATUS = 7

    def __init__(self, name: str, dataset_id: str, client: Client) -> None:
        self._dataset_name = name
        self._dataset_id = dataset_id
        self._client = client

    def upload_description(
        self,
        description: Optional[str] = None,
        collection_time: Optional[str] = None,
        collection_location: Optional[str] = None,
    ) -> None:
        """Upload description of the dataset.

        :param description: description of the dataset to upload
        :param collection_time: collected time of the dataset to upload
        :param collection_location: collected location of the dataset to upload
        """
        post_data = {}
        if description:
            post_data["desc"] = description
        if collection_time:
            post_data["collectedAt"] = collection_time
        if collection_location:
            post_data["collectedLocation"] = collection_location

        if not post_data:
            return

        post_data["contentSetId"] = self._dataset_id
        self._client.dataset_post("updateContentSet", post_data)

    def get_dataset_id(self) -> str:
        """Get the id of this dataset."""
        return self._dataset_id

    def publish(self, version: Optional[str] = None) -> None:
        """Publish a new version of a dataset."""
        post_data = {"contentSetId": self._dataset_id}
        if version:
            post_data["version"] = version
        self._client.dataset_post("publishContentSet", post_data)

    def is_published(self) -> bool:
        """Gheck whether the dataset is published.

        :return: Return true if the dataset is publish, viceversa.
        """
        post_data = {"contentSetId": self._dataset_id}
        data = self._client.dataset_post("listContentSets", post_data)
        status = data["contentSets"][0]["contentSetResp"]["status"]
        return status == DatasetClientBase._PUBLISH_STATUS  # type: ignore[no-any-return]

    def list_segments(self) -> List[str]:
        """List all segment names in a dataset.

        :return: A list of segment names
        """
        return self._list_segments()

    def _list_segments(self, segment_names: Optional[List[str]] = None) -> List[str]:
        post_data: Dict[str, Any] = {"contentSetId": self._dataset_id, "pageSize": -1}
        if segment_names:
            post_data["names"] = segment_names
        segments_info = self._client.dataset_post("listSegments", post_data)["segments"]
        return [SegmentNameWrapper.get_local_segment_name(info["name"]) for info in segments_info]

    def delete_segments(
        self,
        segment_names: Union[str, List[str]],
        force_delete: bool = False,
    ) -> None:
        """Delete segments according to the name list.

        :param name: A single segment Name or a list of the segment names, if empty, the default
        segment will be deleted, if you want to delete all segment, "_all" should be submitted.
        :param force_delete: By default, only segment with no sensor can be deleted.
        If force_delete is true, then sensor and its objects will also be deleted
        """
        if not isinstance(segment_names, list):
            segment_names = [segment_names]
        post_data = {
            "contentSetId": self._dataset_id,
            "names": segment_names,
            "forceDelete": force_delete,
        }
        self._client.dataset_post("deleteSegments", post_data)


class DatasetClient(DatasetClientBase):
    """dataset has only one sensor, supporting create segment."""

    def get_or_create_segment(self, name: str = "") -> DataSegmentClient:
        """Create a segment set according to its name.

        :param name: Segment name, can not be "_default"
        :return: Created segment with given name, if not given name, Created default segment
        """
        if name and not self._list_segments([name]):
            post_data = {"contentSetId": self._dataset_id, "name": name}
            self._client.dataset_post("createOrUpdateSegment", post_data)
        return DataSegmentClient(name, self._dataset_name, self._dataset_id, self._client)

    def get_segment(self, name: str = "") -> DataSegmentClient:
        """Get a segment according to its name.

        :param name: The name of the desired segment
        :raises GASSegmentError: When the required segment does not exist
        :return: The desired segment
        """
        if not self._list_segments([name]):
            raise GASSegmentError(name)
        return DataSegmentClient(name, self._dataset_name, self._dataset_id, self._client)

    def upload_segment_object(
        self,
        segment: Segment,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> DataSegmentClient:
        """Upload a `Segment` to the dataset,
        This function will upload all info contains in the input `Segment` object,
        which includes:
        - Create a segment using the name of input `Segment`
        - Upload all `Data` in the Segment to the dataset

        :param segment: The `Segment` object contains the information needs to be upload
        :param jobs: The number of the max workers in multithread upload
        :param skip_uploaded_files: Set it to `True` to skip the uploaded files
        :return: The client used for uploading the data in the `Segment`
        """
        segment_client = self.get_or_create_segment(segment.name)
        segment_filter: Iterable[Data]
        if skip_uploaded_files:
            done_set = set(segment_client.list_data())
            segment_filter = filter(lambda data: data.remote_path not in done_set, segment)
        else:
            segment_filter = segment

        multithread_upload(segment_client.upload_data_object, segment_filter, jobs=jobs)
        return segment_client


class FusionDatasetClient(DatasetClientBase):
    """Client for fusion dataset which has multiple sensors,
    supporting create fusion segment.
    """

    def get_or_create_segment(self, name: str = "") -> FusionDataSegmentClient:
        """Create a fusion segment set according to the given name.

        :param name: Segment name, can not be "_default"
        :return: Created fusion segment with given name, if not given name, Created default segment
        """
        if name and not self._list_segments([name]):
            post_data = {"contentSetId": self._dataset_id, "name": name}
            self._client.dataset_post("createOrUpdateSegment", post_data)
        return FusionDataSegmentClient(name, self._dataset_name, self._dataset_id, self._client)

    def get_segment(self, name: str = "") -> FusionDataSegmentClient:
        """Get a fusion segment according to its name.

        :param name: The name of the desired fusion segment
        :raises GASSegmentError: When the required fusion segment does not exist
        :return: The desired fusion segment
        """
        if self._list_segments([name]):
            return FusionDataSegmentClient(name, self._dataset_name, self._dataset_id, self._client)
        raise GASSegmentError(name)

    def upload_segment_object(
        self,
        segment: FusionSegment,
        *,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
    ) -> FusionDataSegmentClient:
        """Upload a `FusionSegment` to the dataset,
        This function will upload all info contains in the input `FusionSegment` object,
        which includes:
        - Create a segment using the name of input `FusionSegment`
        - Upload all `Sensor` in the segment to the dataset
        - Upload all `Frame` in the segment to the dataset

        :param segment: The `Segment` object needs to be uploaded
        :param jobs: The number of the max workers in multithread upload
        :param skip_uploaded_files: Set it to `True` to skip the uploaded files
        :return: The client used for uploading the data in the `FusionSegment`
        """
        segment_client = self.get_or_create_segment(segment.name)
        sensors = segment.get_sensors()
        for sensor in sensors.values():
            segment_client.upload_sensor_object(sensor)

        segment_filter: Iterable[Tuple[int, Frame]]
        if skip_uploaded_files:
            segment_filter = self._existing_frame_generator(segment_client, segment)
        else:
            segment_filter = enumerate(segment)

        multithread_upload(
            lambda args: segment_client.upload_frame_object(args[1], args[0]),
            segment_filter,
            jobs=jobs,
        )

        return segment_client

    @staticmethod
    def _existing_frame_generator(
        segment_client: FusionDataSegmentClient,
        segment: FusionSegment,
    ) -> Generator[Tuple[int, Frame], None, None]:
        done_frame_dict = (
            segment_client._list_frame_objects_dict()  # pylint: disable=protected-access
        )
        for i, frame in enumerate(segment):
            if i in done_frame_dict:
                remote_frame, frame_id = done_frame_dict[i]
                if len(remote_frame) == len(frame):
                    continue

                segment_client._delete_frames(frame_id)  # pylint: disable=protected-access

            yield i, frame



# !/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""logging utility functions for dumping http request and response"""

import json
from typing import MutableMapping

from requests.models import PreparedRequest, Response
from requests_toolbelt.multipart.encoder import FileWrapper, MultipartEncoder

REQUEST_TEMPLATE = """
===================================================================
########################## HTTP Request ###########################
"url": {}
"method": {}
"headers": {}
"body": {}
"""

RESPONSE_TEMPLATE = """
########################## HTTP Response ##########################
"url": {}
"status_code": {}
"reason": {}
"headers": {}
"content": {}
===================================================================
"""


def dump_request_and_response(response: Response) -> str:
    """dump http request and response

    :param response: http response and response
    :return: string of http request and response for logging, example:
    ===================================================================
    ########################## HTTP Request ###########################
    "url": https://gas.graviti.cn/gatewayv2/content-store/putObject
    "method": POST
    "headers": {
      "User-Agent": "python-requests/2.23.0",
      "Accept-Encoding": "gzip, deflate",
      "Accept": "*/*",
      "Connection": "keep-alive",
      "X-Token": "c3b1808b21024eb38f066809431e5bb9",
      "Content-Type": "multipart/form-data; boundary=5adff1fc0524465593d6a9ad68aad7f9",
      "Content-Length": "330001"
    }
    "body":
    --5adff1fc0524465593d6a9ad68aad7f9
    b'Content-Disposition: form-data; name="contentSetId"\r\n\r\n'
    b'e6110ff1-9e7c-4c98-aaf9-5e35522969b9'

    --5adff1fc0524465593d6a9ad68aad7f9
    b'Content-Disposition: form-data; name="filePath"\r\n\r\n'
    b'4.jpg'

    --5adff1fc0524465593d6a9ad68aad7f9
    b'Content-Disposition: form-data; name="fileData"; filename="4.jpg"\r\n\r\n'
    [329633 bytes of object data]

    --5adff1fc0524465593d6a9ad68aad7f9--

    ########################## HTTP Response ###########
    "url": https://gas.graviti.cn/gatewayv2/content-stor
    "status_code": 200
    "reason": OK
    "headers": {
      "Date": "Sat, 23 May 2020 13:05:09 GMT",
      "Content-Type": "application/json;charset=utf-8",
      "Content-Length": "69",
      "Connection": "keep-alive",
      "Access-Control-Allow-Origin": "*",
      "X-Kong-Upstream-Latency": "180",
      "X-Kong-Proxy-Latency": "112",
      "Via": "kong/2.0.4"
    }
    "content": {
      "success": true,
      "code": "DATACENTER-0",
      "message": "success",
      "data": {}
    }
    ====================================================
    """
    return _dump_request(response.request) + _dump_response(response)


def _dump_request(request: PreparedRequest) -> str:
    """dump http request

    :param request: http request
    :return: string of http request for logging
    """
    headers = _dump_headers(request.headers)
    body = "N/A"
    content_type = request.headers.get("Content-Type", None)
    if content_type:
        if content_type.startswith("multipart/form-data"):
            body = _dump_multipart_encoder(request.body)
        elif isinstance(request.body, bytes):
            body = request.body.decode("unicode_escape")

    return REQUEST_TEMPLATE.format(request.url, request.method, headers, body)


def _dump_response(response: Response) -> str:
    """dump http response

    :param response: http response
    :return: string of http response for logging
    """
    headers = _dump_headers(response.headers)
    content = "N/A"

    content_type = response.headers.get("Content-Type", None)
    if content_type:
        if content_type.startswith("application/json"):
            content = json.dumps(response.json(), indent=2)
        elif content_type.startswith("text"):
            content = response.text
        elif len(response.content) > 512:
            content = f"[{len(response.content)} bytes of object data]"
        else:
            content = str(response.content)

    return RESPONSE_TEMPLATE.format(
        response.url, response.status_code, response.reason, headers, content
    )


def _dump_multipart_encoder(body: MultipartEncoder) -> str:
    """dump MultipartEncoder multipart/form-data Content-Type post_data

    :param body: MultipartEncoder multipart/formdata post data
    :return: string of multipart/form-data post data for logging
    """
    lines = []
    boundary = body.boundary
    for part in body.parts:
        lines.append("\n" + boundary)
        lines.append(str(part.headers))
        body = part.body
        if isinstance(body, FileWrapper):
            if body.fd.closed:
                lines.append("[file closed]")
            else:
                lines.append(f"[{body.fd.tell()} bytes of object data]")
        else:
            lines.append(str(body.getvalue()))

    lines.append("\n" + boundary + "--")

    return "\n".join(lines)


def _dump_headers(headers: MutableMapping[str, str]) -> str:
    """dump http headers as json format string for logging

    :param headers: headers of http request or response
    :return: json format string of headers content
    """
    return json.dumps(dict(headers), indent=2)



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class TypeEnum and TypeClass"""

from enum import Enum
from typing import Any, Dict, Optional, Type


class TypeEnum(Enum):
    """This is a superclass for Enum which needs creating a mapping with class.
    The 'type' property is used for get the corresponding class of the Enum
    """

    __registry__: Dict["TypeEnum", Type[Any]] = {}

    def __init_subclass__(cls) -> None:
        cls.__registry__ = {}

    @property
    def type(self) -> Type[Any]:
        """Get the corresponding class"""
        return self.__registry__[self]


class TypeClass:  # pylint: disable=too-few-public-methods
    """This is a superclass for the class which needs to link with TypeEnum
    It provides the class variable 'TYPE' to access the corresponding TypeEnum
    """

    _TYPE: Any

    def __init_subclass__(cls, enum: Optional[Any] = None) -> None:
        if not enum:
            return

        cls._TYPE = enum
        enum.__registry__[enum] = cls

    @property
    def enum(self) -> Any:
        """Get the corresponding TypeEnum"""
        return self._TYPE



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

""" label and labeled shape classes """

from .label import (
    Classification,
    Label,
    LabeledBox2D,
    LabeledBox3D,
    LabeledPolygon2D,
    LabeledPolyline2D,
    LabelType,
)
from .label_table import AttributeType, CategoryInfo, LabelTable, LabelTables

__all__ = [
    "Classification",
    "Label",
    "LabelTable",
    "LabelTables",
    "LabelType",
    "LabeledBox2D",
    "LabeledBox3D",
    "LabeledPolygon2D",
    "LabeledPolyline2D",
    "AttributeType",
    "CategoryInfo",
]



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Frame."""

from typing import Any, Dict, Iterator, MutableMapping, Optional, Sequence

from ..geometry import Quaternion, Transform3D
from .data import Data


class Frame(MutableMapping[str, Data]):
    """This class defines the concept of frame,
    which represents a series of Data corresponding to different sensors.

    :param loads: {
        "pose": {
            "translation": {
                "x":
                "y":
                "z":
            },
            "rotation": {
                "w":
                "x":
                "y":
                "z":
            },
        },
        "frame": {
            <key>: data_dict{...},
            <key>: data_dict{...},
            ...
            ...
        }
    }
    """

    def __init__(self, *, loads: Optional[Dict[str, Any]] = None) -> None:
        self._data: Dict[str, Data] = {}
        self._pose = None

        if loads:
            for sensor, data_dict in loads["frame"].items():
                self._data[sensor] = Data(loads=data_dict)
            if "pose" in loads:
                self._pose = Transform3D(loads=loads["pose"])

    def dumps(self) -> Dict[str, Any]:
        """dump a Frame into a frame_dict"""

        frame_dict: Dict[str, Any] = {}
        if self._pose:
            frame_dict["pose"] = self._pose.dumps()

        content_dict = {}
        for sensor, data in self._data.items():
            content_dict[sensor] = data.dumps()
        frame_dict["frame"] = content_dict

        return frame_dict

    def __getitem__(self, key: str) -> Data:
        return self._data.__getitem__(key)

    def __setitem__(self, key: str, value: Data) -> None:
        self._data.__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        self._data.__delitem__(key)

    def __len__(self) -> int:
        return self._data.__len__()

    def __iter__(self) -> Iterator[str]:
        return self._data.__iter__()

    def __repr__(self) -> str:
        str_list = [f"{self.__class__.__name__}{{"]
        for key, value in self._data.items():
            str_list.append(f'  "{key}": {value},')
        str_list.append("}")

        return "\n".join(str_list)

    @property
    def pose(self) -> Optional[Transform3D]:
        """Get the pose of the frame.

        :return: A Transform3D class object representing the pose of the frame
        """
        return self._pose

    def set_pose(
        self,
        transform: Transform3D.TransformType = None,
        *,
        translation: Optional[Sequence[float]] = None,
        rotation: Quaternion.ArgsType = None,
        loads: Optional[Dict[str, Dict[str, float]]] = None,
        **kwargs: Quaternion.KwargsType,
    ) -> None:
        """Set the pose of the current frame
        :param transform: A Transform3D object or a 4x4 or 3x4 transfrom matrix
        :param translation: Translation in a sequence of [x, y, z]
        :param rotation: Rotation in a sequence of [w, x, y, z] or 3x3 rotation matrix or Quaternion
        :param loads: A dictionary containing the translation and the rotation
        :param kwargs: Other parameters to initialize rotation of the transform
        """
        self._pose = Transform3D(
            transform, translation=translation, rotation=rotation, loads=loads, **kwargs
        )



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class PointList2D, Polygon2D"""

from typing import (
    Any,
    Dict,
    Iterable,
    List,
    MutableSequence,
    Optional,
    Sequence,
    TypeVar,
    Union,
    overload,
)

from .box import Box2D
from .vector import Vector2D

T = TypeVar("T", bound="PointList2D")  # pylint: disable=invalid-name


class PointList2D(MutableSequence[Vector2D]):
    """this class defines the concept of point list

    :param points: a list of 2D point list
    :param loads: a list of 2D point dict
    """

    def __init__(
        self,
        points: Optional[Sequence[Sequence[float]]] = None,
        *,
        loads: Optional[List[Dict[str, Any]]] = None,
    ) -> None:
        self._data: List[Vector2D] = []
        if loads:
            for point_dict in loads:
                self._data.append(Vector2D(loads=point_dict))
            return

        if not points:
            return

        for point in points:
            self._data.append(Vector2D(point))

    def dumps(self) -> List[Dict[str, float]]:
        """dump a PointList2D into a point list."""

        return [point.dumps() for point in self._data]

    def __repr__(self) -> str:
        str_list = [f"{self.__class__.__name__}["]
        for point in self._data:
            str_list.append(f"  ({point.x}, {point.y}),")
        str_list.append("]")

        return "\n".join(str_list)

    def insert(self, index: int, value: Vector2D) -> None:
        """Insert Vector2D before index."""
        return self._data.insert(index, value)

    @overload
    def __getitem__(self, index: int) -> Vector2D:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[Vector2D]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[List[Vector2D], Vector2D]:
        return self._data.__getitem__(index)

    @overload
    def __setitem__(self, index: int, value: Vector2D) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[Vector2D]) -> None:
        ...

    def __setitem__(
        self, index: Union[int, slice], value: Union[Vector2D, Iterable[Vector2D]]
    ) -> None:
        self._data.__setitem__(index, value)  # type: ignore[index, assignment]

    def __delitem__(self, index: Union[int, slice]) -> None:
        self._data.__delitem__(index)

    def __len__(self) -> int:
        return self._data.__len__()

    def bounds(self) -> Box2D:
        """calculate the bounds of point list."""

        x_min = x_max = self._data[0].x
        y_min = y_max = self._data[0].y

        for point in self._data:
            if point.x < x_min:
                x_min = point.x
            elif point.x > x_max:
                x_max = point.x

            if point.y < y_min:
                y_min = point.y
            elif point.y > y_max:
                y_max = point.y

        return Box2D(x_min, y_min, x_max, y_max)


class Polygon2D(PointList2D):
    """this class defines the concept of Polygon2D based on class PointList2D"""

    def area(self) -> float:
        """Get the area of the polygon.
        If the orientation of the points is counterclockwise, the area is positive.
        If it is clockwise, the area is negative.

        :return: The area of the polygon
        """
        area = 0.0
        for i in range(len(self._data)):
            # pylint: disable=invalid-name
            x1, y1 = self._data[i - 1]
            x2, y2 = self._data[i]
            area += x1 * y2 - x2 * y1
        return area / 2



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class CatagoryInfo, AttributeInfo and LabelTable."""

from enum import Enum
from typing import (
    Any,
    Dict,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Mapping,
    Optional,
    Type,
    Union,
    ValuesView,
)

from ..label import LabelType
from ..utility import NameClass, NameSortedDict


class AttributeType(Enum):
    """All the possible type of the attributes."""

    boolean = bool
    integer = int
    number = float
    string = str
    array = list
    null = None
    instance = "instance"


class CategoryInfo(NameClass):
    """Information of a category, includes category name and description

    :param name: The name of the category
    :param loads: A dict contains all information of the category
    :raises
        TypeError: Name is required when not given loads
    """


class AttributeInfo(NameClass):
    """Information of a attribute

    :param name: The name of the attribute
    :param enum: All the possible values of the attribute
    :param attribute_type: The type of the attribute value
    :param minimum: The minimum value of number type attribute
    :param maximum: The maximum value of number type attribute
    :is_array: Whether the attribute is a sequence of values or not
    :param parent_categories: The parent categories of the attribute
    :param loads: A dict contains all information of the attribute
    :raises
        TypeError: Name and values are required when not given loads
    """

    SingleArgType = Union[
        str,
        None,
        Type[Any],
        AttributeType,
    ]
    ArgType = Union[SingleArgType, Iterable[SingleArgType]]

    def __init__(
        self,
        name: Optional[str] = None,
        *,
        enum: Optional[Iterable[str]] = None,
        attribute_type: Optional[ArgType] = None,
        minimum: Optional[float] = None,
        maximum: Optional[float] = None,
        is_array: bool = False,
        parent_categories: Union[None, str, Iterable[str]] = None,
        loads: Optional[Dict[str, Any]] = None,
    ):
        if loads:
            super().__init__(loads=loads)
            self._parent_categories = loads.get("parentCategories", [])
            if loads.get("type", None) == "array":
                self._is_array = True
                schema = loads["items"]
            else:
                self._is_array = False
                schema = loads

            self._enum = schema.get("enum", None)
            self._type = self._convert_type(schema.get("type", None))
            self._minimum = schema.get("minimum", None)
            self._maximum = schema.get("maximum", None)
            return

        super().__init__(name)

        self._enum = enum
        self._type = self._convert_type(attribute_type)
        self._minimum = minimum
        self._maximum = maximum
        self._is_array = is_array

        if not parent_categories:
            self._parent_categories = []
        elif isinstance(parent_categories, str):
            self._parent_categories = [parent_categories]
        else:
            self._parent_categories = list(parent_categories)

    def _convert_type(self, type_: ArgType) -> Union[AttributeType, List[AttributeType], None]:
        if not type_:
            return None

        if isinstance(type_, Iterable) and not isinstance(type_, str):  # pylint: disable=W1116
            return [self._convert_single_type(single_type) for single_type in type_]

        return self._convert_single_type(type_)

    @staticmethod
    def _convert_single_type(type_: Union[str, Type[Any], AttributeType, None]) -> AttributeType:
        if isinstance(type_, str):
            return AttributeType[type_]

        if isinstance(type_, AttributeType):
            return type_

        return AttributeType(type_)

    def _dump_type(self) -> Union[str, List[str], None]:
        if isinstance(self._type, list):
            return [type_.name for type_ in self._type]

        if self._type:
            return self._type.name

        return None

    def dumps(self) -> Dict[str, Any]:
        """Dumps the information of this attribute as a dictionary.

        :return: A dictionary contains all information of this attribute
        """
        schema: Dict[str, Any] = {}
        if self._type:
            schema["type"] = self._dump_type()
        if self._enum:
            schema["enum"] = self._enum
        if self._minimum is not None:
            schema["minimum"] = self._minimum
        if self._maximum is not None:
            schema["maximum"] = self._maximum

        data: Dict[str, Any] = super().dumps()
        if self._is_array:
            data["type"] = "array"
            data["items"] = schema
        else:
            data.update(schema)
        if self._parent_categories:
            data["parentCategories"] = self._parent_categories
        return data


class LabelTable:
    """A table contains all labels in a specific label type

    :param is_tracking: A boolean value indicates whether corresponding table is tracking related
    :param loads: A dict contains all information of the label table
    """

    def __init__(
        self, is_tracking: bool = False, *, loads: Optional[Dict[str, Any]] = None
    ) -> None:
        self._categories: NameSortedDict[CategoryInfo] = NameSortedDict()
        self._attributes: NameSortedDict[AttributeInfo] = NameSortedDict()

        if loads:
            self._is_tracking: bool = loads.get("isTracking", False)
            self.description: Optional[str] = loads.get("description", None)
            for category in loads.get("categories", []):
                self.add_category(loads=category)
            for attribute in loads.get("attributes", []):
                self.add_attribute(loads=attribute)
            return

        self._is_tracking = is_tracking
        self.description = None

    def __repr__(self) -> str:
        str_list = [f"{self.__class__.__name__}["]

        for category in self._categories.values():
            str_list.append(f"  {str(category)},")

        for attribute in self._attributes.values():
            str_list.append(f"  {str(attribute)},")

        str_list.append("]")

        return "\n".join(str_list)

    @property
    def is_tracking(self) -> bool:
        """Return whether the label is a tracking label

        :return: If it this a tracking label, return `True`, otherwise return `False`
        """
        return self._is_tracking

    @property
    def categories(self) -> NameSortedDict[CategoryInfo]:
        """Return all the categories of the LabelTable.

        :return: A NameSortedDict of all the categories
        """
        return self._categories

    def dumps(self) -> Dict[str, Any]:
        """Dumps the information of this LabelTable into a dictionary.

        :return: A dictionary contains all information of this LabelTable
        """
        data: Dict[str, Any] = {}

        categories = [category.dumps() for category in self._categories.values()]
        attributes = [attribute.dumps() for attribute in self._attributes.values()]

        if self._is_tracking:
            data["isTracking"] = self._is_tracking
        if self.description:
            data["description"] = self.description
        if categories:
            data["categories"] = categories
        if attributes:
            data["attributes"] = attributes

        return data

    def add_category(
        self, name: Optional[str] = None, *, loads: Optional[Dict[str, Any]] = None
    ) -> None:
        """Add a category to the LabelTable

        :param name: The name of the category
        :param loads: A dict contains all information of the category
        """
        self._categories.add(CategoryInfo(name, loads=loads))

    def add_attribute(
        self,
        name: Optional[str] = None,
        *,
        enum: Optional[Iterable[str]] = None,
        attribute_type: Optional[AttributeInfo.ArgType] = None,
        minimum: Optional[float] = None,
        maximum: Optional[float] = None,
        is_array: bool = False,
        parent_categories: Union[None, str, Iterable[str]] = None,
        loads: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a attribute to the LabelTable

        :param name: The name of the attribute
        :param enum: All the possible values of the attribute
        :param attribute_type: The type of the attribute value
        :param minimum: The minimum value of number type attribute
        :param maximum: The maximum value of number type attribute
        :is_array: Whether the attribute is a sequence of values or not
        :param parent_categories: The parent categories of the attribute
        :param loads: A dict contains all information of the attribute
        """
        self._attributes.add(
            AttributeInfo(
                name,
                enum=enum,
                attribute_type=attribute_type,
                minimum=minimum,
                maximum=maximum,
                is_array=is_array,
                parent_categories=parent_categories,
                loads=loads,
            )
        )


class LabelTables(Mapping[LabelType, LabelTable]):
    """Label tables is a mapping which contains `LabelTable`,
    the corresponding key is the 'name' of `LabelType`.

    :param loads: A dict contains a series of label tabel dicts

    """

    def __init__(self, *, loads: Optional[Dict[str, Dict[str, Any]]] = None) -> None:
        self._data: Dict[LabelType, LabelTable] = {}

        if loads:
            for type_name, label_table in loads.items():
                self.create_label_table(LabelType[type_name], loads=label_table)
            return

    def __getitem__(self, key: LabelType) -> LabelTable:
        return self._data.__getitem__(key)

    def __len__(self) -> int:
        return self._data.__len__()

    def __iter__(self) -> Iterator[LabelType]:
        return self._data.__iter__()

    def __repr__(self) -> str:
        str_list = [f"{self.__class__.__name__}{{"]

        for label_type in self._data:
            str_list.append(f"  {label_type.name},")

        str_list.append("}")
        return "\n".join(str_list)

    def create_label_table(
        self,
        label_type: LabelType,
        is_tracking: bool = False,
        *,
        loads: Optional[Dict[str, Any]] = None,
    ) -> LabelTable:
        """Create a new label table and add it to label tables.

        :param label_type: the label type of the label table
        :param is_tracking: A boolean value indicates whether corresponding task
                            is tracking related.
        :param loads: A dict contains all information of the label table

        :return: created new label table
        """

        label_table = LabelTable(is_tracking, loads=loads)
        self._data[label_type] = label_table
        return label_table

    def keys(self) -> KeysView[LabelType]:
        """Return keys view of the label tables' keys

        :return: keys view
        """
        return self._data.keys()

    def values(self) -> ValuesView[LabelTable]:
        """Return values view of the label tables' values.

        :return: values view
        """
        return self._data.values()

    def items(self) -> ItemsView[LabelType, LabelTable]:
        """Return items view of the label tables' items.

        :return: items view
        """
        return self._data.items()

    def dumps(self) -> Dict[str, Any]:
        """Dump the label tables into a series of label table dict.

        :return: a dict contains a series of label table dict with their label types as dict keys
        """
        label_table_dicts: Dict[str, Any] = {}
        for label_type, label_table in self._data.items():
            label_table_dicts[label_type.name] = label_table.dumps()
        return label_table_dicts



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Data"""

import os
from http.client import HTTPResponse
from typing import (
    Any,
    Callable,
    Dict,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Mapping,
    Optional,
    Union,
    ValuesView,
    overload,
)
from urllib.request import urlopen

from _io import BufferedReader
from typing_extensions import Literal

from ..label import (
    Classification,
    Label,
    LabeledBox2D,
    LabeledBox3D,
    LabeledPolygon2D,
    LabeledPolyline2D,
    LabelType,
)
from ..utility import TBRN

Labels = Union[
    Classification,
    List[LabeledBox2D],
    List[LabeledBox3D],
    List[LabeledPolygon2D],
    List[LabeledPolyline2D],
]


class Data(Mapping[LabelType, Labels]):
    """this class defines the concept of labels to one file

    :param fileuri: the file path of the labeled object in this data
    :param timestamp: the timestamp when collecting the labeled object in this data
    :param loads: {
        "fileuri": <str>,
        "timestamp": <float>,
        "labels_classification": {...},
        "labels_box2D": {...},
        "labels_box3D": {...},
        "labels_polygon": {...},
        "labels_polyline": {...},
    }
    :TypeError: when fileuri is not given
    """

    def __init__(
        self,
        fileuri: Union[str, TBRN, None] = None,
        *,
        remote_path: Optional[str] = None,
        timestamp: Optional[float] = None,
        url_getter: Optional[Callable[[TBRN], str]] = None,
        loads: Optional[Dict[str, Any]] = None,
    ) -> None:

        self._labels: Dict[LabelType, Any] = {}
        self._remote_path = remote_path  # The remote storage location of the data
        self._url_getter = url_getter

        if loads:
            fileuri = loads["fileuri"]
            timestamp = loads.get("timestamp", None)
            self.load_labels(loads)

        if not fileuri:
            raise TypeError(
                f"{self.__class__.__name__}() missing 1 required positional argument: 'fileuri'"
            )

        if isinstance(fileuri, str) and fileuri.startswith("tb:"):
            fileuri = TBRN(tbrn=fileuri)

        self._fileuri = fileuri
        self._timestamp = timestamp

        self._local_path = self._fileuri if isinstance(self._fileuri, str) else ""

    def dumps(self) -> Dict[str, Any]:
        """dump a data into a dict"""

        data_dict: Dict[str, Any] = {}
        data_dict["fileuri"] = str(self._fileuri)
        if self._timestamp is not None:
            data_dict["timestamp"] = self._timestamp

        data_dict.update(self.dump_labels())

        return data_dict

    def dump_labels(self) -> Dict[str, Any]:
        """dump all labels into a dict"""

        labels_dict: Dict[str, Any] = {}
        for key, labels in self._labels.items():
            if key == LabelType.CLASSIFICATION:
                labels_dict[key.value] = labels.dumps()
            else:
                labels_dict[key.value] = [label.dumps() for label in labels]

        return labels_dict

    def load_labels(self, loads: Dict[str, Any]) -> None:
        """load all labels into the data object"""

        for label_type in LabelType:
            if label_type.value not in loads:
                continue

            labels = loads[label_type.value]

            if label_type == LabelType.CLASSIFICATION:
                self._labels[label_type] = label_type.type(loads=labels)
            else:
                self._labels[label_type] = [label_type.type(loads=label) for label in labels]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self._fileuri}")'

    def get_url(self) -> str:
        """Return the url of the data host by tensorbay

        :return: the url of this data
        :raises ValueError: when the url_getter is missing
        """
        if not self._url_getter:
            raise ValueError(
                f"{self.__repr__()} has no url_getter, it is probably not a remote file"
            )

        return self._url_getter(self.tbrn)

    def open(self) -> Union[BufferedReader, HTTPResponse]:
        """Return binary fp of this file, the file pointer will be get by buildin `open()` for
        local file, by `urllib.request.urlopen()` for remote file

        :return: The file pointer for this data, the retuen class will be `_io.BufferedReader` for
        local file, `http.client.HTTPResponse` for remote file

        """
        if self._local_path:
            return open(self._local_path, "rb")

        return urlopen(self.get_url())

    @property
    def fileuri(self) -> Union[str, TBRN]:
        """Return fileuri of the data.

        :return: fileuri of the data
        """
        return self._fileuri

    @property
    def timestamp(self) -> Optional[float]:
        """Return timestamp of the data.

        :return: timestamp of the data
        """
        return self._timestamp

    @property
    def tbrn(self) -> TBRN:
        """Return timestamp of the data.

        :return: timestamp of the data
        :raises ValueError: when the tbrn is missing
        """
        if not isinstance(self._fileuri, TBRN):
            raise ValueError(f"{self.__repr__()} has no TBRN, it is probably not a remote file")

        return self._fileuri

    @property
    def local_path(self) -> str:
        """Return local path of the data.

        :return: local path of the data
        """
        return self._local_path

    @property
    def remote_path(self) -> str:
        """Get the remote path of the data

        :return: the remote path of the data
        """
        if self._remote_path:
            return self._remote_path

        if isinstance(self._fileuri, TBRN):
            return self._fileuri.remote_path

        return os.path.basename(self._fileuri)

    @remote_path.setter
    def remote_path(self, remote_path: str) -> None:
        """Set the remote path of the data """
        self._remote_path = remote_path

    def register_label(self, label_types: Union[LabelType, Iterable[LabelType]]) -> None:
        """register one or several label types to data

        :param label_types: the LabelType to register
        """
        if not isinstance(label_types, Iterable):  # pylint: disable=W1116
            label_types = [label_types]

        for label_type in label_types:
            if label_type == LabelType.CLASSIFICATION:
                continue

            self._labels[label_type] = []

    def append_label(self, label: Label) -> None:
        """append a label to labels

        :param label: a Label to append
        :raises TypeError: when the label type of the append label is not registerd
        """

        if label.enum == LabelType.CLASSIFICATION:
            self._labels[label.enum] = label
            return

        try:
            self._labels[label.enum].append(label)
        except KeyError as error:
            raise TypeError(
                "LabelType needs to be registerd before appending label "
                f"(call register_label(LabelType.{label.enum.name}))"
            ) from error

    def __len__(self) -> int:
        return self._labels.__len__()

    def __iter__(self) -> Iterator[LabelType]:
        return self._labels.__iter__()

    @overload
    def __getitem__(self, label_type: Literal[LabelType.CLASSIFICATION]) -> Classification:
        ...

    @overload
    def __getitem__(self, label_type: Literal[LabelType.BOX2D]) -> List[LabeledBox2D]:
        ...

    @overload
    def __getitem__(self, label_type: Literal[LabelType.BOX3D]) -> List[LabeledBox3D]:
        ...

    @overload
    def __getitem__(self, label_type: Literal[LabelType.POLYGON]) -> List[LabeledPolygon2D]:
        ...

    @overload
    def __getitem__(self, label_type: Literal[LabelType.POLYLINE]) -> List[LabeledPolyline2D]:
        ...

    @overload
    def __getitem__(self, label_type: LabelType) -> Labels:
        ...

    def __getitem__(self, label_type: LabelType) -> Labels:
        """return one type of labels in a data

        :param label_type: required LabelType
        """
        return self._labels[label_type]  # type: ignore[no-any-return]

    def keys(self) -> KeysView[LabelType]:
        """A set-like object providing a view on the data label

        :return: keys view of data label
        """
        return self._labels.keys()

    def values(self) -> ValuesView[Labels]:
        """An object providing a view on the data label values

        :return: values view of the data label
        """
        return self._labels.values()

    def items(self) -> ItemsView[LabelType, Labels]:
        """A set-like object providing a view on the data label items

        :return: items view of the data label
        """
        return self._labels.items()



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class Polyline2D"""

import math
from typing import Any, Dict, Iterator, List, Sequence, Tuple

import numpy as np

from .polygon import PointList2D


class Polyline2D(PointList2D):
    """this class defines the concept of Polyline2D based on class PointList2D"""

    @staticmethod
    def _distance(point1: Sequence[float], point2: Sequence[float]) -> float:
        return math.sqrt((point1[0] - point2[0]) ** 2 + (point1[1] - point2[1]) ** 2)

    @staticmethod
    def _get_polyline_info(polyline: Sequence[Sequence[float]]) -> List[Dict[str, Any]]:
        points = np.array(polyline, dtype=np.float32)
        vector = [points[i] - points[i - 1] for i in range(1, len(points))]
        length = [Polyline2D._distance(v, [0, 0]) for v in vector]
        distance = np.cumsum(length)
        time = distance / distance[-1]
        info: List[Dict[str, Any]] = []

        for index, v in enumerate(vector):  # pylint: disable=invalid-name
            info.append(
                {
                    "vector": v,
                    "last_time": time[index - 1] if index else 0,
                    "time": time[index],
                    "point": points[index],
                    "index": index,
                }
            )
        return info

    @staticmethod
    def _insert_point(info1: Dict[str, Any], info2: Dict[str, Any]) -> Dict[str, Any]:
        """Insert one point in info1 into the info2.

        :param info1: segment info of the insert point
        :param info2: the inserted segment info
        :return: a dictionary containing info of the inserted point
        """
        ratio = (info1["time"] - info2["last_time"]) / (info2["time"] - info2["last_time"])
        insert_point = info2["point"] + info2["vector"] * ratio
        return {"index": info2["index"] + 1, "point": insert_point}

    @staticmethod
    def _insert_points(
        polyline_info1: Iterator[Dict[str, Any]],
        polyline_info2: Iterator[Dict[str, Any]],
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        """Calculate insert points in polyline1 and polyline2.

        :param polyline_info1: segment info in polyline1
        :param polyline_info2: segment info in polyline2
        :returns:
            insert points in polyline1
            insert points in polyline2
        """
        insert_points1: List[Dict[str, Any]] = []
        insert_points2: List[Dict[str, Any]] = []
        info1 = next(polyline_info1)
        info2 = next(polyline_info2)

        try:
            while True:
                if info1["time"] < info2["time"]:
                    insert_points2.append(Polyline2D._insert_point(info1, info2))
                    info1 = next(polyline_info1)
                    continue

                if info1["time"] > info2["time"]:
                    insert_points1.append(
                        Polyline2D._insert_point(  # pylint: disable=arguments-out-of-order
                            info2, info1
                        )
                    )
                    info2 = next(polyline_info2)
                    continue

                if info1["time"] == info2["time"]:
                    info1 = next(polyline_info1)
                    info2 = next(polyline_info2)
        except StopIteration:
            pass
        return insert_points1, insert_points2

    @staticmethod
    def _max_distance_in_point_pairs(polyline1: np.ndarray, polyline2: np.ndarray) -> float:
        """Calculate the maximum distance between point pairs in two polylines.

        :param polyline1: the first inserted polyline
        :param polyline2: the second inserted polyline, containing the same number of
            points as the first one
        :return: the maximum distance between point pairs in two polylines
        """
        assert len(polyline1) == len(polyline2)

        max_distance = -1.0
        for point1, point2 in zip(polyline1, polyline2):
            distance = Polyline2D._distance(point1, point2)
            if distance > max_distance:
                max_distance = distance
        return max_distance

    @staticmethod
    def uniform_frechet_distance(
        polyline1: Sequence[Sequence[float]],
        polyline2: Sequence[Sequence[float]],
    ) -> float:
        """Compute the maximum distance between two curves if walk on a constant speed on a curve.

        :param polyline1: the first polyline consisting of multiple points
        :param polyline2: the second polyline consisting of multiple points
        :return: the computed distance between the two polylines
        """
        polyline_info1 = Polyline2D._get_polyline_info(polyline1)
        polyline_info2 = Polyline2D._get_polyline_info(polyline2)
        line2_reverse = list(reversed(polyline2))
        polyline_info2_reverse = Polyline2D._get_polyline_info(line2_reverse)

        # forward
        insert_points1, insert_points2 = Polyline2D._insert_points(
            iter(polyline_info1), iter(polyline_info2)
        )
        line1 = polyline1
        line2 = polyline2
        for point in reversed(insert_points1):
            line1 = np.insert(line1, point["index"], point["point"], axis=0)
        for point in reversed(insert_points2):
            line2 = np.insert(line2, point["index"], point["point"], axis=0)
        distance_forward = Polyline2D._max_distance_in_point_pairs(line1, line2)

        # reverse
        insert_points1, insert_points2 = Polyline2D._insert_points(
            iter(polyline_info1), iter(polyline_info2_reverse)
        )
        for point in reversed(insert_points1):
            polyline1 = np.insert(polyline1, point["index"], point["point"], axis=0)
        for point in reversed(insert_points2):
            line2_reverse = np.insert(line2_reverse, point["index"], point["point"], axis=0)
        distance_reverse = Polyline2D._max_distance_in_point_pairs(polyline1, line2_reverse)

        return min(distance_forward, distance_reverse)

    @staticmethod
    def similarity(
        polyline1: Sequence[Sequence[float]],
        polyline2: Sequence[Sequence[float]],
    ) -> float:
        """Calculate the similarity between two polylines, range from 0 to 1.

        :param polyline1: the first polyline consisting of multiple points
        :param polyline2: the second polyline consisting of multiple points
        :return: the similarity between the two polylines. The bigger, the more similar.
        """
        min_distance = Polyline2D.uniform_frechet_distance(polyline1, polyline2)
        max_distance = -1.0
        for point1 in polyline1:
            for point2 in polyline2:
                distance = Polyline2D._distance(point1, point2)
                if distance > max_distance:
                    max_distance = distance
        return 1 - min_distance / max_distance



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file defines class NameClass"""

from typing import (
    Dict,
    ItemsView,
    Iterator,
    KeysView,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
    ValuesView,
    overload,
)

from sortedcontainers import SortedDict


class NameClass:
    """A mixin class for instance which has immutable name and mutable description

    :param name: A string representing the class name
    :param loads: A dict contains name and description
    :raises
        TypeError: Name required when not given loads
    """

    def __init__(
        self,
        name: Optional[str] = None,
        *,
        loads: Optional[Dict[str, str]] = None,
    ) -> None:
        if loads:
            self._name = loads["name"]
            self.description = loads.get("description", "")
            return

        if name is None:
            raise TypeError(
                f"{self.__class__.__name__}() missing 1 required positional argument: 'name'."
            )

        self._name = name
        self.description = ""

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self._name}")'

    def dumps(self) -> Dict[str, str]:
        """Dumps the instance as a dictionary.

        :return: A dictionary containing name and description
        """
        data = {"name": self._name}
        if self.description:
            data["description"] = self.description

        return data

    @property
    def name(self) -> str:
        """Return name of the instance.

        :return: Name of the instance
        """
        return self._name


T = TypeVar("T", bound=NameClass)  # pylint: disable=invalid-name


class NameSortedDict(Mapping[str, T]):
    """Name sorted dict is a sorted mapping which contains `NameClass`,
    the corrsponding key is the 'name' of `NameClass`.
    Name sorted dict keys are maintained in sorted order.

    :param data: A mapping from str to `NameClass` which needs to be transferred to `NameSortedDict`
    """

    def __init__(self, data: Optional[Mapping[str, T]] = None) -> None:
        self._data = SortedDict(data)

    def __getitem__(self, key: str) -> T:
        return self._data.__getitem__(key)  # type: ignore[no-any-return]

    def __len__(self) -> int:
        return self._data.__len__()  # type: ignore[no-any-return]

    def __iter__(self) -> Iterator[str]:
        return self._data.__iter__()  # type: ignore[no-any-return]

    def __repr__(self) -> str:
        str_list = ["{"]
        for key, value in self._data.items():
            str_list.append(f'  "{key}": {value.__class__.__name__}(),')
        str_list.append("}")
        return "\n".join(str_list)

    def add(self, value: T) -> None:
        """Store item in name sorted dict"""
        self._data[value.name] = value

    def keys(self) -> KeysView[str]:
        """Return new name sorted keys view of the sorted dict's keys.

        :return: new name sorted keys view
        """
        return self._data.keys()  # type: ignore[no-any-return]

    def values(self) -> ValuesView[T]:
        """Return new name sorted values view of the sorted dict's values.

        :return: new name sorted values view
        """
        return self._data.values()  # type: ignore[no-any-return]

    def items(self) -> ItemsView[str, T]:
        """Return new name sorted items view of the sorted dict's items.

        :return: new name sorted items view
        """
        return self._data.items()  # type: ignore[no-any-return]


class NameSortedList(Sequence[T]):
    """Name sorted list is a sorted sequence which contains `NameClass`,
    Name sorted list are maintained in sorted order according to the name of NameClass.
    """

    def __init__(self) -> None:
        self._data = SortedDict()

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[T]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[T, List[T]]:
        return self._data.values()[index]  # type: ignore[no-any-return]

    def __len__(self) -> int:
        return self._data.__len__()  # type: ignore[no-any-return]

    def __iter__(self) -> Iterator[T]:
        return self._data.values().__iter__()  # type: ignore[no-any-return]

    def __repr__(self) -> str:
        str_list = ["["]
        for value in self._data.values():
            str_list.append(f"  {value},")
        str_list.append("]")
        return "\n".join(str_list)

    def add(self, value: T) -> None:
        """Store item in name sorted list."""
        self._data[value.name] = value

    def get_from_name(self, name: str) -> T:
        """Get item in name sorted list from name of NameClass."""
        return self._data[name]  # type: ignore[no-any-return]
